import asyncio as aio
import logging
import uuid

from ble2mqtt.devices.base import LIGHT_DOMAIN, ConnectionMode, Device
from ble2mqtt.devices.uuids import FIRMWARE_VERSION
from ble2mqtt.protocols.avea import AveaProtocol

AVEA_CONTROL = uuid.UUID("f815e811-456c-6761-746f-4d756e696368")

_LOGGER = logging.getLogger(__name__)

LIGHT_ENTITY = 'light'


class AveaBulb(AveaProtocol, Device):
    NAME = 'avea_rgbw'
    DATA_CHAR = AVEA_CONTROL
    ACTIVE_SLEEP_INTERVAL = 1
    RECONNECTION_SLEEP_INTERVAL = 30
    MANUFACTURER = 'Avea'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    SEND_DATA_PERIOD = 60

    def __init__(self, mac, *args, **kwargs):
        super().__init__(mac, *args, **kwargs)
        self._color = (255, 255, 255)  # rgb, if OFF store previous color
        self._real_color = (255, 255, 255)  # rgb
        self._brightness = 255

        self.initial_status_sent = False

    @property
    def entities(self):
        return {
            LIGHT_DOMAIN: [
                {
                    'name': 'light',
                },
            ],
        }

    def get_values_by_entities(self):
        return {
            LIGHT_ENTITY: {
                'state': (
                    'ON'
                    if (
                        self._brightness != 0 and
                        self._real_color != (0, 0, 0)
                    )
                    else 'OFF'
                ),
                'brightness': self._brightness,
                'color': {
                    'r': self._color[0],
                    'g': self._color[1],
                    'b': self._color[2],
                },
                'color_mode': 'rgb',
            },
        }

    async def get_device_data(self):
        await super().get_device_data()
        try:
            self._model = await self.read_name()
        except Exception as e:
            _LOGGER.warning(f'[{self}] Cannot read name: {e}')
            self._model = 'Bulb'
        version = await self._read_with_timeout(FIRMWARE_VERSION)
        if isinstance(version, (bytes, bytearray)):
            self._version = version.decode(errors='ignore').strip('\0')

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        timer = 0
        while True:
            await self.update_device_data(send_config)
            await self.read_state()

            if not self.initial_status_sent or timer >= self.SEND_DATA_PERIOD:
                await self._notify_state(publish_topic)
                timer = 0
            timer += self.ACTIVE_SLEEP_INTERVAL
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)

    async def handle_messages(self, publish_topic, *args, **kwargs):
        while True:
            try:
                if not self.client.is_connected:
                    raise ConnectionError()
                message = await aio.wait_for(
                    self.message_queue.get(),
                    timeout=60,
                )
            except aio.TimeoutError:
                await aio.sleep(1)
                continue
            value = message['value']
            entity_topic, action_postfix = self.get_entity_subtopic_from_topic(
                message['topic'],
            )
            if entity_topic == self._get_topic_for_entity(
                self.get_entity_by_name(LIGHT_DOMAIN, LIGHT_ENTITY),
                skip_unique_id=True,
            ):
                _LOGGER.info(f'[{self}] set light {value}')

                if value.get('brightness'):
                    self._brightness = value['brightness']
                    await self.write_brightness(self._brightness)

                target_color = None
                if value.get('color'):
                    color = value['color']
                    try:
                        target_color = (
                            color['r'],
                            color['g'],
                            color['b'],
                        )
                    except ValueError:
                        return
                    if target_color != (0, 0, 0):
                        self._color = target_color

                if value.get('state'):
                    state = self.transform_value(value['state'])
                    if state == 'ON' and target_color is None:
                        target_color = self._color
                    if state == 'OFF':
                        target_color = (0, 0, 0)
                if target_color is not None:
                    await self.write_color(*target_color)
                await self._notify_state(publish_topic)

    def handle_color(self, value):
        self._real_color = value
        if value != (0, 0, 0):
            self._color = value

    def handle_brightness(self, value):
        self._brightness = value



import asyncio as aio
import logging
import uuid
from contextlib import asynccontextmanager

from ..protocols.redmond import (COOKER_PREDEFINED_PROGRAMS, CookerRunState,
                                 CookerState, RedmondCookerProtocol,
                                 RedmondError)
from .base import (SELECT_DOMAIN, SENSOR_DOMAIN, SWITCH_DOMAIN, ConnectionMode,
                   Device)
from .uuids import DEVICE_NAME

_LOGGER = logging.getLogger(__name__)

UUID_NORDIC_TX = uuid.UUID("6e400002-b5a3-f393-e0a9-e50e24dcca9e")
UUID_NORDIC_RX = uuid.UUID("6e400003-b5a3-f393-e0a9-e50e24dcca9e")

COOK_ENTITY = 'cook'
LOCK_ENTITY = 'lock'
SOUND_ENTITY = 'sound'
PREDEFINED_PROGRAM_ENTITY = 'predefined_program'
TEMPERATURE_ENTITY = 'temperature'
MODE_ENTITY = 'mode'


def option_to_const(option):
    return option.replace(' ', '_').lower()


def const_to_option(const):
    return const.replace('_', ' ').title()


class RedmondCooker(RedmondCookerProtocol, Device):
    MAC_TYPE = 'random'
    NAME = 'redmond_rmc_m200'
    TX_CHAR = UUID_NORDIC_TX
    RX_CHAR = UUID_NORDIC_RX
    ACTIVE_SLEEP_INTERVAL = 1
    RECONNECTION_SLEEP_INTERVAL = 30
    MANUFACTURER = 'Redmond'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    SEND_DATA_PERIOD = 5  # seconds when boiling
    STANDBY_SEND_DATA_PERIOD_MULTIPLIER = 12  # 12 * 5 seconds in standby mode

    def __init__(self, mac, key='ffffffffffffffff', default_program='express',
                 *args, loop, **kwargs):
        super().__init__(mac, *args, loop=loop, **kwargs)
        assert isinstance(key, str) and len(key) == 16
        self._default_program = default_program
        self._key = bytes.fromhex(key)
        self._state: CookerState = None

        self._send_data_period_multiplier = \
            self.STANDBY_SEND_DATA_PERIOD_MULTIPLIER
        self.initial_status_sent = False

    @property
    def entities(self):
        return {
            SWITCH_DOMAIN: [
                {
                    'name': COOK_ENTITY,
                    'topic': COOK_ENTITY,
                    'icon': 'pot-steam',
                },
                {
                    'name': LOCK_ENTITY,
                    'topic': LOCK_ENTITY,
                    'icon': 'lock',
                },
                {
                    'name': SOUND_ENTITY,
                    'topic': SOUND_ENTITY,
                    'icon': 'music-note',
                },
            ],
            SENSOR_DOMAIN: [
                {
                    'name': TEMPERATURE_ENTITY,
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': MODE_ENTITY,
                },
            ],
            SELECT_DOMAIN: [
                {
                    'name': PREDEFINED_PROGRAM_ENTITY,
                    'topic': PREDEFINED_PROGRAM_ENTITY,
                    'options': [
                        const_to_option(x)
                        for x in COOKER_PREDEFINED_PROGRAMS.keys()
                    ],
                },
            ],
        }

    def get_values_by_entities(self):
        return {
            TEMPERATURE_ENTITY: self._state.target_temperature,
            MODE_ENTITY: self._state.state.name.title().replace('_', ' '),
            PREDEFINED_PROGRAM_ENTITY: const_to_option(
                {
                    (state.program, state.subprogram): k
                    for k, state in COOKER_PREDEFINED_PROGRAMS.items()
                }.get(
                    (self._state.program, self._state.subprogram),
                    '',
                ),
            ),
        }

    async def update_device_state(self):
        state = await self.get_mode()
        if state:
            self._state = state
            self.update_multiplier()
            self.initial_status_sent = False

    async def get_device_data(self):
        await self.protocol_start()
        await self.login(self._key)
        model = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(model, (bytes, bytearray)):
            self._model = model.decode()
        else:
            # macos can't access characteristic
            self._model = 'RMC'
        version = await self.get_version()
        if version:
            self._version = f'{version[0]}.{version[1]}'
        await self.update_device_state()
        # await self.set_time()
        # await self._update_statistics()

    def update_multiplier(self, state: CookerState = None):
        if state is None:
            state = self._state
        self._send_data_period_multiplier = (
            1
            if state.state in [
                CookerRunState.HEAT,
                CookerRunState.SETUP_PROGRAM,
                CookerRunState.WARM_UP,
                CookerRunState.COOKING,
            ]
            else self.STANDBY_SEND_DATA_PERIOD_MULTIPLIER
        )

    async def notify_run_state(self, new_state: CookerState, publish_topic):
        if not self.initial_status_sent or \
                new_state.state != self._state.state:
            state_to_str = {
                True: 'ON',
                False: 'OFF',
            }
            mode = state_to_str[new_state.state not in (
                CookerRunState.OFF,
                CookerRunState.SETUP_PROGRAM,
            )]
            await aio.gather(
                publish_topic(topic=self._get_topic(COOK_ENTITY), value=mode),
                self._notify_state(publish_topic),
            )
            self.initial_status_sent = True
            if self._state != new_state:
                self._state = new_state
                await self._notify_state(publish_topic)
            else:
                self._state = new_state
        else:
            self._state = new_state
        self.update_multiplier()

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        counter = 0
        while True:
            await self.update_device_data(send_config)
            # if boiling notify every 5 seconds, 60 sec otherwise
            new_state = await self.get_mode()
            await self.notify_run_state(new_state, publish_topic)
            counter += 1

            if counter > (
                    self.SEND_DATA_PERIOD * self._send_data_period_multiplier
            ):
                # await self._update_statistics()
                await self._notify_state(publish_topic)
                counter = 0
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)

    async def switch_running_mode(self, value):
        if value == 'ON':
            # switch to SETUP_PROGRAM mode if it is off
            if self._state.state == CookerRunState.OFF:
                try:
                    await self.set_predefined_program(self._default_program)
                except KeyError:
                    # if incorrect program passed, use initial mode
                    await self.set_mode(self._state)
            await self.run()
            next_state = CookerRunState.COOKING  # any of heating stages
        else:
            await self.stop()
            next_state = CookerRunState.OFF
        self.update_multiplier(CookerState(state=next_state))

    async def handle_messages(self, publish_topic, *args, **kwargs):
        def is_entity_topic(entity, topic):
            return topic == self._get_topic_for_entity(
                entity,
                skip_unique_id=True,
            )

        @asynccontextmanager
        async def process_entity_change(entity, value):
            value = self.transform_value(value)
            _LOGGER.info(
                f'[{self}] switch cooker {entity["name"]} value={value}',
            )
            for _ in range(10):
                try:
                    yield
                    # update state to real values
                    await self.update_device_state()
                    await aio.gather(
                        publish_topic(
                            topic=self._get_topic_for_entity(entity),
                            value=self.transform_value(value),
                        ),
                        self._notify_state(publish_topic),
                        loop=self._loop,
                    )
                    break
                except ConnectionError as e:
                    _LOGGER.exception(str(e))
                await aio.sleep(5)

        while True:
            try:
                if not self.client.is_connected:
                    raise ConnectionError()
                message = await aio.wait_for(
                    self.message_queue.get(),
                    timeout=60,
                )
            except aio.TimeoutError:
                await aio.sleep(1)
                continue
            value = message['value']

            entity_topic, action_postfix = self.get_entity_subtopic_from_topic(
                message['topic'],
            )
            entity = self.get_entity_by_name(SWITCH_DOMAIN, COOK_ENTITY)
            if is_entity_topic(entity, entity_topic):
                async with process_entity_change(entity, value):
                    try:
                        await self.switch_running_mode(value)
                    except RedmondError:
                        _LOGGER.exception(
                            f'[{self}] Problem with switching cooker',
                        )
                continue

            entity = self.get_entity_by_name(SWITCH_DOMAIN, SOUND_ENTITY)
            if is_entity_topic(entity, entity_topic):
                async with process_entity_change(entity, value):
                    await self.set_sound(value == 'ON')
                continue

            entity = self.get_entity_by_name(SWITCH_DOMAIN, LOCK_ENTITY)
            if is_entity_topic(entity, entity_topic):
                async with process_entity_change(entity, value):
                    await self.set_lock(value == 'ON')
                continue

            entity = self.get_entity_by_name(
                SELECT_DOMAIN,
                PREDEFINED_PROGRAM_ENTITY,
            )
            if is_entity_topic(entity, entity_topic):
                try:
                    value = option_to_const(value)
                except KeyError:
                    _LOGGER.error(f'{self} program "{value}" does not exist')
                    continue
                _LOGGER.info(f'set predefined program {value}')

                while True:
                    try:
                        await self.set_predefined_program(value)
                        # update state to real values
                        await self.update_device_state()
                        await aio.gather(
                            publish_topic(
                                topic=self._get_topic_for_entity(entity),
                                value=const_to_option(
                                    self.transform_value(value),
                                ),
                            ),
                            self._notify_state(publish_topic),
                            loop=self._loop,
                        )
                        break
                    except ConnectionError as e:
                        _LOGGER.exception(str(e))
                    await aio.sleep(5)
                continue



import logging
from dataclasses import dataclass

from bleak.backends.device import BLEDevice

from ..utils import format_binary
from .uuids import ENVIRONMENTAL_SENSING
from .xiaomi_base import XiaomiHumidityTemperature

_LOGGER = logging.getLogger(__name__)


@dataclass
class SensorState:
    battery: int = 0
    temperature: float = 0
    humidity: float = 0
    sends_custom: bool = False  # support decimal value for humidity


class XiaomiHumidityTemperatureLYWSDATC(XiaomiHumidityTemperature):
    NAME = 'xiaomilywsd_atc'
    SENSOR_CLASS = SensorState
    SUPPORT_PASSIVE = True
    SUPPORT_ACTIVE = False

    def handle_advert(self, scanned_device: BLEDevice, adv_data):
        service_data = adv_data.service_data
        adv_data = service_data.get(str(ENVIRONMENTAL_SENSING))

        if adv_data:
            sends_custom = bool(self._state and self._state.sends_custom)
            if len(adv_data) == 15:
                # b'\xe6o\xb98\xc1\xa4\x95\t\xff\x08~\x0cd\xe0\x04'
                sends_custom = True
                adv_data = bytes(adv_data)
                self._state = self.SENSOR_CLASS(
                    temperature=int.from_bytes(
                        adv_data[6:8], byteorder='little', signed=True) / 100,
                    humidity=int.from_bytes(
                        adv_data[8:10], byteorder='little') / 100,
                    battery=adv_data[12],
                    sends_custom=sends_custom,
                )

            elif len(adv_data) == 13:
                if sends_custom:
                    # ignore low res humidity packets
                    return
                # [a4 c1 38 84 7e 97 01 26 15 50 0b 73 17]
                #  <----- mac -----> temp hum bat
                adv_data = bytes(adv_data)
                self._state = self.SENSOR_CLASS(
                    temperature=int.from_bytes(
                        adv_data[6:8], byteorder='big', signed=True) / 10,
                    humidity=adv_data[8],
                    battery=adv_data[9],
                    sends_custom=sends_custom,
                )
                _LOGGER.debug(
                    f'Advert received for {self}, {format_binary(adv_data)}, '
                    f'current state: {self._state}',
                )



import asyncio as aio
import json
import logging
import typing as ty
from contextlib import asynccontextmanager
from uuid import getnode

import aio_mqtt
from bleak import BleakError, BleakScanner
from bleak.backends.device import BLEDevice

from .devices.base import (BINARY_SENSOR_DOMAIN, CLIMATE_DOMAIN, COVER_DOMAIN,
                           DEVICE_TRACKER_DOMAIN, LIGHT_DOMAIN, SELECT_DOMAIN,
                           SENSOR_DOMAIN, SWITCH_DOMAIN, ConnectionMode,
                           ConnectionTimeoutError, Device, done_callback)

_LOGGER = logging.getLogger(__name__)

CONFIG_MQTT_NAMESPACE = 'homeassistant'
BRIDGE_STATE_TOPIC = 'state'
BLUETOOTH_ERROR_RECONNECTION_TIMEOUT = 60
FAILURE_LIMIT = 5


ListOfConnectionErrors = (
    BleakError,
    aio.TimeoutError,

    # dbus-next exceptions:
    # AttributeError: 'NoneType' object has no attribute 'call'
    AttributeError,
    # https://github.com/hbldh/bleak/issues/409
    EOFError,
)


# initialize in a loop
BLUETOOTH_RESTARTING: aio.Lock = None  # type: ignore


async def run_tasks_and_cancel_on_first_return(*tasks: aio.Future,
                                               return_when=aio.FIRST_COMPLETED,
                                               ignore_futures=(),
                                               ) -> ty.Sequence[aio.Future]:
    async def cancel_tasks(_tasks) -> ty.List[aio.Task]:
        # cancel first, then await. Because other tasks can raise exceptions
        # while switching tasks
        canceled = []
        for t in _tasks:
            if t in ignore_futures:
                continue
            if not t.done():
                t.cancel()
                canceled.append(t)
        tasks_raise_exceptions = []
        for t in canceled:
            try:
                await t
            except aio.CancelledError:
                pass
            except Exception:
                _LOGGER.exception(
                    f'Unexpected exception while cancelling tasks! {t}',
                )
                tasks_raise_exceptions.append(t)
        return tasks_raise_exceptions

    assert all(isinstance(t, aio.Future) for t in tasks)
    try:
        # NB: pending tasks can still raise exception or finish
        # while tasks are switching
        done, pending = await aio.wait(tasks, return_when=return_when)
    except aio.CancelledError:
        await cancel_tasks(tasks)
        # it could happen that tasks raised exception and canceling wait task
        # abandons tasks with exception
        for t in tasks:
            if not t.done() or t.cancelled():
                continue
            try:
                t.result()
            # no CancelledError expected
            except Exception:
                _LOGGER.exception(
                    f'Task raises exception while cancelling parent coroutine '
                    f'that waits for it {t}')
        raise

    # while switching tasks for await other pending tasks can raise an exception
    # we need to append more tasks to the result if so
    await cancel_tasks(pending)

    task_remains = [t for t in pending if not t.cancelled()]
    return [*done, *task_remains]


async def handle_returned_tasks(*tasks: aio.Future):
    raised = [t for t in tasks if t.done() and t.exception()]
    returned_normally = set(tasks) - set(raised)

    results = []

    if raised:
        task_for_raise = raised.pop()
        for t in raised:
            try:
                await t
            except aio.CancelledError:
                raise
            except Exception:
                _LOGGER.exception('Task raised an error')
        await task_for_raise
    for t in returned_normally:
        results.append(await t)
    return results


def hardware_exception_occurred(exception):
    ex_str = str(exception)
    return (
        'org.freedesktop.DBus.Error.ServiceUnknown' in ex_str or
        'org.freedesktop.DBus.Error.NoReply' in ex_str or
        'org.freedesktop.DBus.Error.AccessDenied' in ex_str or
        'org.bluez.Error.Failed: Connection aborted' in ex_str or
        'org.bluez.Error.NotReady' in ex_str or
        'org.bluez.Error.InProgress' in ex_str
    )


ListOfMQTTConnectionErrors = (
        aio_mqtt.ConnectionLostError,
        aio_mqtt.ConnectionClosedError,
        aio_mqtt.ServerDiedError,
        BrokenPipeError,
)


async def restart_bluetooth():
    if BLUETOOTH_RESTARTING.locked():
        await aio.sleep(9)
        return
    async with BLUETOOTH_RESTARTING:
        _LOGGER.warning('Restarting bluetoothd...')
        proc = await aio.create_subprocess_exec(
            'hciconfig', 'hci0', 'down',
        )
        await proc.wait()
        proc = await aio.create_subprocess_exec(
            '/etc/init.d/bluetoothd', 'restart',
        )
        await proc.wait()
        await aio.sleep(3)
        proc = await aio.create_subprocess_exec(
            'hciconfig', 'hci0', 'up',
        )
        await proc.wait()
        await aio.sleep(5)
        _LOGGER.warning('Restarting bluetoothd finished')


@asynccontextmanager
async def handle_ble_exceptions():
    try:
        yield
    except ListOfConnectionErrors as e:
        if hardware_exception_occurred(e):
            await restart_bluetooth()
            await aio.sleep(3)
        raise


class DeviceManager:
    def __init__(self, device, *, mqtt_client, base_topic, config_prefix,
                 global_availability_topic):
        self.device: Device = device
        self._mqtt_client = mqtt_client
        self._base_topic = base_topic
        self._config_prefix = config_prefix
        self._global_availability_topic = global_availability_topic
        self.manage_task = None

    async def close(self):
        if self.manage_task and not self.manage_task.done():
            self.manage_task.cancel()
            try:
                await self.manage_task
            except aio.CancelledError:
                pass
        self.manage_task = None
        try:
            await self.device.close()
        except aio.CancelledError:
            raise
        except Exception:
            _LOGGER.exception(f'Problem on closing device {self.device}')

    def run_task(self) -> aio.Task:
        assert not self.manage_task, \
            f'{self.device} Previous task was not finished! {self.manage_task}'
        self.manage_task = aio.create_task(self.manage_device())
        return self.manage_task

    async def publish_topic_callback(self, topic, value, nowait=False):
        _LOGGER.debug(f'call publish callback topic={topic} value={value}')
        if not self._mqtt_client.is_connected():
            _LOGGER.warning(f'{self.device} mqtt is disconnected')
            return
        await self._mqtt_client.publish(
            aio_mqtt.PublishableMessage(
                topic_name='/'.join((self._base_topic, topic)),
                payload=value,
                qos=aio_mqtt.QOSLevel.QOS_1,
            ),
            nowait=nowait,
        )

    def _get_topic(self, dev_id, subtopic, *args):
        return '/'.join(
            filter(None, (self._base_topic, dev_id, subtopic, *args)),
        )

    @property
    def _config_device_topic(self):
        """Add a prefix to avoid interfering with other ble software"""

        return f'{self._config_prefix}{self.device.dev_id}'

    async def send_device_config(self):
        device = self.device
        device_info = {
            'identifiers': [
                device.unique_id,
            ],
            'name': device.unique_name,
            'model': device.model,
        }
        if device.manufacturer:
            device_info['manufacturer'] = device.manufacturer
        if device.version:
            device_info['sw_version'] = device.version

        def get_generic_vals(entity: dict):
            name = entity.pop('name')
            result = {
                'name': f'{name}_{device.friendly_id}',
                'unique_id': f'{name}_{device.dev_id}',
                'device': device_info,
                'availability_mode': 'all',
                'availability': [
                    {'topic': self._global_availability_topic},
                    {'topic': '/'.join(
                        (self._base_topic, self.device.availability_topic),
                    )},
                ],
            }
            icon = entity.pop('icon', None)
            if icon:
                result['icon'] = f'mdi:{icon}'
            entity.pop('topic', None)
            entity.pop('json', None)
            entity.pop('main_value', None)
            result.update(entity)
            return result

        messages_to_send = []
        for cls, entities in device.entities_with_lqi.items():
            if cls in (
                BINARY_SENSOR_DOMAIN,
                SENSOR_DOMAIN,
                DEVICE_TRACKER_DOMAIN,
            ):
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    if entity.get('json') and entity.get('main_value'):
                        state_topic_part = {
                            'json_attributes_topic': state_topic,
                            'state_topic': state_topic,
                            'value_template':
                                f'{{{{ value_json.{entity["main_value"]} }}}}',
                        }
                    else:
                        state_topic_part = {
                            'state_topic': state_topic,
                            'value_template':
                                f'{{{{ value_json.{entity_name} }}}}',
                        }

                    if cls == DEVICE_TRACKER_DOMAIN:
                        state_topic_part['source_type'] = 'bluetooth_le'

                    payload = json.dumps({
                        **get_generic_vals(entity),
                        **state_topic_part,
                    })
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == SWITCH_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    command_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    payload = json.dumps({
                        **get_generic_vals(entity),
                        'state_topic': state_topic,
                        'command_topic': command_topic,
                    })
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
                    # TODO: send real state on receiving status from a device
                    _LOGGER.debug(f'Publish initial state topic={state_topic}')
                    await self._mqtt_client.publish(
                        aio_mqtt.PublishableMessage(
                            topic_name=state_topic,
                            payload='OFF',
                            qos=aio_mqtt.QOSLevel.QOS_1,
                        ),
                    )
            if cls == LIGHT_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    set_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    payload = json.dumps({
                        **get_generic_vals(entity),
                        'schema': 'json',
                        'color_mode': bool(entity.get('color_mode', True)),
                        'supported_color_modes': entity.get(
                            'color_mode',
                            ['rgb'],
                        ),
                        'brightness': entity.get('brightness', True),
                        'state_topic': state_topic,
                        'command_topic': set_topic,
                    })
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == COVER_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    set_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    set_position_topic = '/'.join(
                        (state_topic, device.SET_POSITION_POSTFIX),
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    config_params = {
                        **get_generic_vals(entity),
                        'state_topic': state_topic,
                        'position_topic': state_topic,
                        'json_attributes_topic': state_topic,
                        'value_template': '{{ value_json.state }}',
                        'position_template': '{{ value_json.position }}',
                        'command_topic': set_topic,
                        'set_position_topic': set_position_topic,
                    }
                    payload = json.dumps(config_params)
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == SELECT_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    set_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    config_params = {
                        **get_generic_vals(entity),
                        'state_topic': state_topic,
                        'command_topic': set_topic,
                    }
                    payload = json.dumps(config_params)
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == CLIMATE_DOMAIN:
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', device.STATE_TOPIC),
                    )
                    mode_command_topic = '/'.join(
                        (state_topic, device.SET_MODE_POSTFIX),
                    )
                    temperature_command_topic = '/'.join(
                        (state_topic, device.SET_TARGET_TEMPERATURE_POSTFIX),
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        self._config_device_topic,
                        entity_name,
                        'config',
                    ))
                    config_params = {
                        **get_generic_vals(entity),
                        'current_temperature_topic': state_topic,
                        'current_temperature_template':
                            '{{ value_json.temperature }}',
                        'mode_state_topic': state_topic,
                        'mode_state_template': '{{ value_json.mode }}',
                        'mode_command_topic': mode_command_topic,
                        'temperature_state_topic': state_topic,
                        'temperature_state_template':
                            '{{ value_json.target_temperature }}',
                        'temperature_command_topic': temperature_command_topic,
                        'json_attributes_topic': state_topic,
                        'temp_step': 0.5,
                    }
                    payload = json.dumps(config_params)
                    _LOGGER.debug(
                        f'Publish config topic={config_topic}: {payload}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
        await aio.gather(*[
            self._mqtt_client.publish(message)
            for message in messages_to_send
        ])
        device.config_sent = True

    async def send_availability(self, value: bool):
        return await self.device.send_availability(
            self.publish_topic_callback,
            value,
        )

    async def _sleep_until_next_connection(self):
        device = self.device
        _LOGGER.debug(
            f'Sleep for {device.RECONNECTION_SLEEP_INTERVAL} secs to '
            f'reconnect to device={device}',
        )
        if device._connection_mode == ConnectionMode.ACTIVE_KEEP_CONNECTION:
            try:
                await aio.wait_for(
                    device._advertisement_seen.wait(),
                    timeout=device.RECONNECTION_SLEEP_INTERVAL,
                )
            except aio.TimeoutError:
                pass
        else:
            await aio.sleep(self.device.RECONNECTION_SLEEP_INTERVAL)

    async def publish_topic_with_availability(self, topic, value):
        # call sequentially to allow HA receive a new value
        await self.publish_topic_callback(topic, value)
        await self.send_availability(True)

    async def manage_device(self):
        device = self.device
        _LOGGER.debug(f'Start managing device={device}')
        failure_count = 0
        missing_device_count = 0
        while True:
            async with BLUETOOTH_RESTARTING:
                _LOGGER.debug(f'[{device}] Check for lock')
            try:
                async with handle_ble_exceptions():
                    await device.connect()
                    initial_coros = []
                    if not device.is_passive:
                        if not device.DEVICE_DROPS_CONNECTION:
                            initial_coros.append(device.disconnected_event.wait)
                        await device.get_device_data()
                        failure_count = 0
                        missing_device_count = 0

                    if device.subscribed_topics:
                        await self._mqtt_client.subscribe(*[
                            (
                                '/'.join((self._base_topic, topic)),
                                aio_mqtt.QOSLevel.QOS_1,
                            )
                            for topic in device.subscribed_topics
                        ])
                    _LOGGER.debug(f'[{device}] mqtt subscribed')
                    coros = [
                        *[coro() for coro in initial_coros],
                        device.handle(
                            self.publish_topic_with_availability,
                            send_config=self.send_device_config,
                        ),
                    ]
                    will_handle_messages = bool(device.subscribed_topics)
                    if will_handle_messages:
                        coros.append(
                            device.handle_messages(
                                self.publish_topic_with_availability,
                            ),
                        )

                    tasks = [aio.create_task(t) for t in coros]
                    _LOGGER.debug(f'[{device}] tasks are created')

                    await run_tasks_and_cancel_on_first_return(*tasks)
                    if device.disconnected_event.is_set():
                        _LOGGER.debug(f'{device} has disconnected')
                    finished = [t for t in tasks if not t.cancelled()]
                    await handle_returned_tasks(*finished)
            except aio.CancelledError:
                # the only way to send availability=False on program shutdown
                if self.device.ACTIVE_CONNECTION_MODE in (
                    ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT,
                    ConnectionMode.ON_DEMAND_CONNECTION,
                ):
                    try:
                        await aio.wait_for(
                            self.send_availability(False),
                            timeout=1,
                        )
                    except aio.TimeoutError:
                        pass
                raise
            except KeyboardInterrupt:
                raise
            except ConnectionTimeoutError:
                missing_device_count += 1
                _LOGGER.error(
                    f'[{device}] connection problem, '
                    f'attempts={missing_device_count}',
                )
            except (ConnectionError, TimeoutError, aio.TimeoutError):
                missing_device_count += 1
                _LOGGER.exception(
                    f'[{device}] connection problem, '
                    f'attempts={missing_device_count}',
                )
            except ListOfConnectionErrors as e:
                if 'Device with address' in str(e) and \
                        'was not found' in str(e):
                    missing_device_count += 1
                    _LOGGER.warning(
                        f'Error while connecting to {device}, {e} {repr(e)}, '
                        f'attempts={missing_device_count}',
                    )
                else:
                    # if isinstance(e, aio.TimeoutError) or \
                    #         'org.bluez.Error.Failed: Connection aborted' in \
                    #         str(e):
                    failure_count += 1
                    _LOGGER.warning(
                        f'Error while connecting to {device}, {e} {repr(e)}, '
                        f'failure_count={failure_count}',
                    )

                # sometimes LYWSD03MMC devices remain connected
                # and doesn't advert their presence.
                # If cannot find device for several attempts, restart
                # the bluetooth chip
                if missing_device_count >= device.CONNECTION_FAILURES_LIMIT:
                    _LOGGER.error(
                        f'Device {device} was not found for '
                        f'{missing_device_count} times. Restarting bluetooth.',
                    )
                    missing_device_count = 0
                    await restart_bluetooth()
            finally:
                if self.device.ACTIVE_CONNECTION_MODE not in (
                    ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT,
                    ConnectionMode.ON_DEMAND_CONNECTION,
                ):
                    try:
                        await aio.wait_for(
                            self.send_availability(False),
                            timeout=1,
                        )
                    except aio.TimeoutError:
                        pass
                try:
                    await aio.wait_for(device.close(), timeout=5)
                except aio.CancelledError:
                    raise
                except Exception:
                    _LOGGER.exception(f'{device} problem on device.close()')
                try:
                    canceled = []
                    for t in tasks:
                        if not t.done():
                            t.cancel()
                            canceled.append(t)
                    for t in canceled:
                        try:
                            t.result()
                        except aio.CancelledError:
                            pass
                except aio.CancelledError:
                    raise
                except Exception:
                    pass

            if failure_count >= FAILURE_LIMIT:
                await restart_bluetooth()
                failure_count = 0
            try:
                if not device.disconnected_event.is_set():
                    await aio.wait_for(
                        device.disconnected_event.wait(),
                        timeout=10,
                    )
            except aio.TimeoutError:
                _LOGGER.exception(f'{device} not disconnected in 10 secs')
            await self._sleep_until_next_connection()


class Ble2Mqtt:
    TOPIC_ROOT = 'ble2mqtt'
    BRIDGE_TOPIC = 'bridge'

    def __init__(
            self,
            host: str,
            port: int = None,
            user: ty.Optional[str] = None,
            password: ty.Optional[str] = None,
            reconnection_interval: int = 10,
            loop: ty.Optional[aio.AbstractEventLoop] = None,
            *,
            base_topic,
            mqtt_config_prefix,
    ) -> None:
        global BLUETOOTH_RESTARTING

        self._mqtt_host = host
        self._mqtt_port = port
        self._mqtt_user = user
        self._mqtt_password = password
        self._base_topic = base_topic
        self._mqtt_config_prefix = mqtt_config_prefix

        self._reconnection_interval = reconnection_interval
        self._loop = loop or aio.get_event_loop()
        BLUETOOTH_RESTARTING = aio.Lock()

        self._mqtt_client = aio_mqtt.Client(
            client_id_prefix=f'{base_topic}_',
            loop=self._loop,
        )

        self._device_managers: ty.Dict[Device, DeviceManager] = {}

        self.availability_topic = '/'.join((
            self._base_topic,
            self.BRIDGE_TOPIC,
            BRIDGE_STATE_TOPIC,
        ))

        self.device_registry: ty.List[Device] = []

    async def start(self):
        result = await run_tasks_and_cancel_on_first_return(
            self._loop.create_task(self._connect_forever()),
            self._loop.create_task(self._handle_messages()),
        )
        for t in result:
            await t

    async def close(self) -> None:
        for device, manager in self._device_managers.items():
            await manager.close()

        if self._mqtt_client.is_connected:
            try:
                await self._mqtt_client.disconnect()
            except aio.CancelledError:
                raise
            except Exception as e:
                _LOGGER.warning(f'Error on MQTT  disconnecting: {repr(e)}')

    def register(self, device_class: ty.Type[Device], *args, **kwargs):
        device = device_class(*args, **kwargs)
        if not device:
            return
        if not device.is_passive and not device.SUPPORT_ACTIVE:
            raise NotImplementedError(
                f'Device {device.dev_id} doesn\'t support active mode',
            )
        assert device.is_passive or device.ACTIVE_CONNECTION_MODE in (
            ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT,
            ConnectionMode.ACTIVE_KEEP_CONNECTION,
            ConnectionMode.ON_DEMAND_CONNECTION,
        )
        self.device_registry.append(device)

    @property
    def subscribed_topics(self):
        return [
            '/'.join((self._base_topic, topic))
            for device in self.device_registry
            for topic in device.subscribed_topics
        ]

    async def _handle_messages(self) -> None:
        async for message in self._mqtt_client.delivered_messages(
            f'{self._base_topic}/#',
        ):
            _LOGGER.debug(message)
            while True:
                if message.topic_name not in self.subscribed_topics:
                    await aio.sleep(0)
                    continue

                prefix = f'{self._base_topic}/'
                if message.topic_name.startswith(prefix):
                    topic_wo_prefix = message.topic_name[len(prefix):]
                else:
                    topic_wo_prefix = prefix
                for _device in self.device_registry:
                    if topic_wo_prefix in _device.subscribed_topics:
                        device = _device
                        break
                else:
                    raise NotImplementedError('Unknown topic')
                await aio.sleep(0)
                if not device.client.is_connected:
                    _LOGGER.warning(
                        f'Received topic {topic_wo_prefix} '
                        f'with {message.payload} '
                        f' but {device.client} is offline',
                    )
                    await aio.sleep(5)
                    continue

                try:
                    value = json.loads(message.payload)
                except ValueError:
                    value = message.payload.decode()

                await device.add_incoming_message(topic_wo_prefix, value)
                break

            await aio.sleep(1)

    async def stop_device_manage_tasks(self):
        for manager in self._device_managers.values():
            try:
                await manager.close()
            except aio.CancelledError:
                raise
            except Exception:
                _LOGGER.exception(
                    f'Problem on closing dev manager {manager.device}')

    def device_detection_callback(self, device: BLEDevice, advertisement_data):
        for reg_device in self.device_registry:
            if reg_device.mac.lower() == device.address.lower():
                if device.rssi:
                    # update rssi for all devices if available
                    reg_device.rssi = device.rssi
                if reg_device.is_passive:
                    if device.name:
                        reg_device._model = device.name
                    reg_device.handle_advert(device, advertisement_data)
                else:
                    _LOGGER.debug(
                        f'active device seen: {reg_device} '
                        f'{advertisement_data}',
                    )
                    reg_device.set_advertisement_seen()

    async def scan_devices_task(self):
        empty_scans = 0
        while True:
            # 10 empty scans in a row means that bluetooth restart is required
            if empty_scans >= 10:
                empty_scans = 0
                await restart_bluetooth()

            try:
                async with handle_ble_exceptions():
                    scanner = BleakScanner()
                    scanner.register_detection_callback(
                        self.device_detection_callback,
                    )
                    try:
                        await aio.wait_for(scanner.start(), 10)
                    except aio.TimeoutError:
                        _LOGGER.error('Scanner start failed with timeout')
                    await aio.sleep(3)
                    devices = scanner.discovered_devices
                    await scanner.stop()
                    if not devices:
                        empty_scans += 1
                    else:
                        empty_scans = 0
                    _LOGGER.debug(f'found {len(devices)} devices: {devices}')
            except KeyboardInterrupt:
                raise
            except aio.IncompleteReadError:
                raise
            except ListOfConnectionErrors as e:
                _LOGGER.exception(e)
                empty_scans += 1
            await aio.sleep(1)

    async def _run_device_tasks(self, mqtt_connection_fut: aio.Future) -> None:
        has_passive_devices = False
        for dev in self.device_registry:
            self._device_managers[dev] = \
                DeviceManager(
                    dev,
                    mqtt_client=self._mqtt_client,
                    base_topic=self._base_topic,
                    config_prefix=self._mqtt_config_prefix,
                    global_availability_topic=self.availability_topic,
                )
            if dev.is_passive:
                has_passive_devices = True
        _LOGGER.debug("Wait for network interruptions...")

        device_tasks = [
            manager.run_task()
            for manager in self._device_managers.values()
        ]
        if has_passive_devices:
            scan_task = self._loop.create_task(self.scan_devices_task())
            scan_task.add_done_callback(done_callback)
            device_tasks.append(scan_task)

        futs = [
            mqtt_connection_fut,
            *device_tasks,
        ]

        finished = await run_tasks_and_cancel_on_first_return(
            *futs,
            ignore_futures=[mqtt_connection_fut],
        )

        finished_managers = []
        for d, m in self._device_managers.items():
            if m.manage_task not in finished:
                await m.close()
            else:
                finished_managers.append(m)

        for m in finished_managers:
            await m.close()

        # when mqtt server disconnects, multiple tasks can raise
        # exceptions. We must fetch all of them
        finished = [t for t in futs if t.done() and not t.cancelled()]
        await handle_returned_tasks(*finished)

    async def _connect_forever(self) -> None:
        dev_id = hex(getnode())
        while True:
            try:
                mqtt_connection = await self._mqtt_client.connect(
                    host=self._mqtt_host,
                    port=self._mqtt_port,
                    username=self._mqtt_user,
                    password=self._mqtt_password,
                    client_id=f'ble2mqtt_{dev_id}',
                    will_message=aio_mqtt.PublishableMessage(
                        topic_name=self.availability_topic,
                        payload='offline',
                        qos=aio_mqtt.QOSLevel.QOS_1,
                        retain=True,
                    ),
                )
                _LOGGER.info(f'Connected to {self._mqtt_host}')
                await self._mqtt_client.publish(
                    aio_mqtt.PublishableMessage(
                        topic_name=self.availability_topic,
                        payload='online',
                        qos=aio_mqtt.QOSLevel.QOS_1,
                        retain=True,
                    ),
                )
                await self._run_device_tasks(mqtt_connection.disconnect_reason)
            except (aio.CancelledError, KeyboardInterrupt):
                await self._mqtt_client.publish(
                    aio_mqtt.PublishableMessage(
                        topic_name=self.availability_topic,
                        payload='offline',
                        qos=aio_mqtt.QOSLevel.QOS_0,
                        retain=True,
                    ),
                )
                raise
            except Exception:
                _LOGGER.exception(
                    "Connection lost. Will retry in %d seconds.",
                    self._reconnection_interval,
                )
                try:
                    await self.stop_device_manage_tasks()
                except aio.CancelledError:
                    raise
                except Exception:
                    _LOGGER.exception('Exception in _connect_forever()')
                try:
                    await self._mqtt_client.disconnect()
                except aio.CancelledError:
                    raise
                except Exception:
                    _LOGGER.error('Disconnect from MQTT broker error')
                await aio.sleep(self._reconnection_interval)



import asyncio as aio
import logging
import typing as ty
import uuid
from dataclasses import dataclass
from enum import Enum

from ..protocols.am43 import AM43Protocol
from .base import SENSOR_DOMAIN, BaseCover, ConnectionMode, CoverRunState

_LOGGER = logging.getLogger(__name__)

BATTERY_ENTITY = 'battery'
ILLUMINANCE_ENTITY = 'illuminance'

BLINDS_CONTROL = uuid.UUID("0000fe51-0000-1000-8000-00805f9b34fb")


class MovementType(Enum):
    STOP = 0
    POSITION = 1


@dataclass
class AM43State:
    battery: ty.Optional[int] = None
    position: int = 0
    illuminance: int = 0
    run_state: CoverRunState = CoverRunState.CLOSED
    target_position: ty.Optional[int] = None


class AM43Cover(AM43Protocol, BaseCover):
    NAME = 'am43'
    MANUFACTURER = 'Blind'
    DATA_CHAR = BLINDS_CONTROL
    ACTIVE_SLEEP_INTERVAL = 1
    SEND_DATA_PERIOD = 5
    STANDBY_SEND_DATA_PERIOD_MULTIPLIER = 12 * 5  # 5 minutes
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    # HA notation. We convert value on setting and receiving data
    CLOSED_POSITION = 0
    OPEN_POSITION = 100

    @property
    def entities(self):
        return {
            **super().entities,
            SENSOR_DOMAIN: [
                {
                    'name': BATTERY_ENTITY,
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
                {
                    'name': ILLUMINANCE_ENTITY,
                    'device_class': 'illuminance',
                    'unit_of_measurement': 'lx',
                },
            ],
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._model = 'AM43'
        self._state = AM43State()
        self.initial_status_sent = False

    def get_values_by_entities(self):
        return {
            BATTERY_ENTITY: self._state.battery,
            ILLUMINANCE_ENTITY: self._state.illuminance,
            self.COVER_ENTITY: {
                'state': self._state.run_state.value,
                'position': self._state.position,
            },
        }

    async def get_device_data(self):
        await super().get_device_data()
        await self.client.start_notify(
            self.DATA_CHAR,
            self.notification_callback,
        )
        await self._update_full_state()

    async def _update_running_state(self):
        await self._get_position()

    async def _update_full_state(self):
        await super()._update_full_state()

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        # request every SEND_DATA_PERIOD if running and
        # SEND_DATA_PERIOD * STANDBY_SEND_DATA_PERIOD_MULTIPLIER if in
        # standby mode

        timer = 0
        while True:
            await self.update_device_data(send_config)
            # if running notify every 5 seconds, 60 sec otherwise
            is_running = self._state.run_state in [
                CoverRunState.OPENING,
                CoverRunState.CLOSING,
            ]
            multiplier = (
                1 if is_running else self.STANDBY_SEND_DATA_PERIOD_MULTIPLIER
            )

            timer += self.ACTIVE_SLEEP_INTERVAL
            if not self.initial_status_sent or \
                    timer >= self.SEND_DATA_PERIOD * multiplier:
                if is_running:
                    _LOGGER.debug(f'[{self}] check for position')
                    await self._update_running_state()
                    if self._state.position == self.CLOSED_POSITION:
                        _LOGGER.info(
                            f'[{self}] Minimum position reached. Set to CLOSED',
                        )
                        self._state.run_state = CoverRunState.CLOSED
                    elif self._state.position == self.OPEN_POSITION:
                        _LOGGER.info(
                            f'[{self}] Maximum position reached. Set to OPEN',
                        )
                        self._state.run_state = CoverRunState.OPEN
                else:
                    _LOGGER.debug(f'[{self}] check for full state')
                    await self._update_full_state()
                await self._notify_state(publish_topic)
                timer = 0
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)

    def handle_battery(self, value):
        self._state.battery = value

    def handle_position(self, value):
        self._state.position = value

    def handle_illuminance(self, value):
        self._state.illuminance = value



import abc
import asyncio as aio
import json
import logging
import typing as ty
import uuid
from collections import defaultdict, namedtuple
from enum import Enum

from bleak import BleakClient, BleakError
from bleak.backends.device import BLEDevice

from ..devices.uuids import DEVICE_NAME, FIRMWARE_VERSION
from ..utils import format_binary, rssi_to_linkquality

_LOGGER = logging.getLogger(__name__)
registered_device_types = {}


BINARY_SENSOR_DOMAIN = 'binary_sensor'
CLIMATE_DOMAIN = 'climate'
COVER_DOMAIN = 'cover'
DEVICE_TRACKER_DOMAIN = 'device_tracker'
LIGHT_DOMAIN = 'light'
SELECT_DOMAIN = 'select'
SENSOR_DOMAIN = 'sensor'
SWITCH_DOMAIN = 'switch'

DEFAULT_STATE_TOPIC = ''  # send to the parent topic


class CoverRunState(Enum):
    OPEN = 'open'
    OPENING = 'opening'
    CLOSED = 'closed'
    CLOSING = 'closing'
    STOPPED = 'stopped'


class ConnectionMode(Enum):
    PASSIVE = 0
    ACTIVE_POLL_WITH_DISCONNECT = 1
    ACTIVE_KEEP_CONNECTION = 2
    ON_DEMAND_CONNECTION = 3  # not implemented yet


class ConnectionTimeoutError(ConnectionError):
    pass


def done_callback(future: aio.Future):
    exc_info = None
    try:
        exc_info = future.exception()
    except aio.CancelledError:
        pass

    if exc_info is not None:
        exc_info = (  # type: ignore
            type(exc_info),
            exc_info,
            exc_info.__traceback__,
        )
        _LOGGER.exception(
            f'{future} stopped unexpectedly',
            exc_info=exc_info,
        )


class RegisteredType(abc.ABCMeta):
    def __new__(mcs, clsname, superclasses, attributedict):
        newclass = type.__new__(mcs, clsname, superclasses, attributedict)
        # condition to prevent base class registration
        if superclasses and abc.ABC not in superclasses:
            if newclass.NAME is not None:
                registered_device_types[newclass.NAME] = newclass
            assert (
                not newclass.SUPPORT_ACTIVE or
                newclass.ACTIVE_CONNECTION_MODE is not None
            ), f'{clsname} requires ACTIVE_CONNECTION_MODE to be set'
        return newclass


class BaseDevice(abc.ABC, metaclass=RegisteredType):
    NAME: str = None  # type: ignore
    SUPPORT_PASSIVE: bool = False
    SUPPORT_ACTIVE: bool = True
    ACTIVE_CONNECTION_MODE: ConnectionMode = None  # type: ignore

    # Whether we should stop handle task on disconnect or not
    # if true wait more to publish data to topics
    DEVICE_DROPS_CONNECTION: bool = False

    def __init__(self, *args, loop, **kwargs):
        self._loop = loop
        self.client: BleakClient = None
        self.disconnected_event = aio.Event()
        if kwargs.get('passive') and not self.SUPPORT_PASSIVE:
            raise NotImplementedError(
                'This device doesn\'t support passive mode',
            )
        self._is_passive = kwargs.get('passive', self.SUPPORT_PASSIVE)
        if self._is_passive:
            self._connection_mode: ConnectionMode = ConnectionMode.PASSIVE
        else:
            self._connection_mode = self.ACTIVE_CONNECTION_MODE
        self.config_sent = False

    @property
    def is_passive(self):
        return self._is_passive

    async def close(self):
        pass

    async def _read_with_timeout(self, char, timeout=5):
        try:
            result = await aio.wait_for(
                self.client.read_gatt_char(char),
                timeout=timeout,
                loop=self._loop,
            )
        except (aio.TimeoutError, BleakError, AttributeError):
            _LOGGER.exception(f'Cannot connect to device {self}')
            result = None
        return result

    @staticmethod
    def transform_value(value):
        if isinstance(value, bool):
            return 'ON' if value else 'OFF'
        if not isinstance(value, str):
            return value
        vl = value.lower()
        if vl in ['0', 'off', 'no']:
            return 'OFF'
        elif vl in ['1', 'on', 'yes']:
            return 'ON'
        return value

    @property
    def unique_id(self):
        raise NotImplementedError()

    @property
    def unique_name(self):
        raise NotImplementedError()

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        raise NotImplementedError()

    def handle_advert(self, scanned_device: BLEDevice, adv_data):
        raise NotImplementedError()


class Device(BaseDevice, abc.ABC):
    MQTT_VALUES = None
    SET_POSTFIX: str = 'set'
    SET_POSITION_POSTFIX: str = 'set_position'  # for covers. Consider rework
    SET_MODE_POSTFIX: str = 'set_mode'  # for climate
    SET_TARGET_TEMPERATURE_POSTFIX: str = 'set_temperature'  # for climate
    MAC_TYPE: str = 'public'
    MANUFACTURER: str = None  # type: ignore
    CONNECTION_FAILURES_LIMIT = 100
    RECONNECTION_SLEEP_INTERVAL = 60
    ACTIVE_SLEEP_INTERVAL = 60
    PASSIVE_SLEEP_INTERVAL = 60
    # deprecated
    LINKQUALITY_TOPIC: ty.Optional[str] = None
    STATE_TOPIC: str = DEFAULT_STATE_TOPIC

    # secs to sleep if not connected or no data in passive mode
    NOT_READY_SLEEP_INTERVAL = 5

    def __init__(self, mac, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.message_queue: aio.Queue = aio.Queue()
        self.mac = mac
        self.friendly_name = kwargs.pop('friendly_name', None)
        self._model = None
        self._version = None
        self._manufacturer = self.MANUFACTURER
        self._rssi = None
        self._advertisement_seen = aio.Event()

        assert set(self.entities.keys()) <= {
            BINARY_SENSOR_DOMAIN,
            CLIMATE_DOMAIN,
            COVER_DOMAIN,
            DEVICE_TRACKER_DOMAIN,
            LIGHT_DOMAIN,
            SELECT_DOMAIN,
            SENSOR_DOMAIN,
            SWITCH_DOMAIN,
        }, f'Unknown domain: {list(self.entities.keys())}'

    def set_advertisement_seen(self):
        self._advertisement_seen.set()

    def _get_topic(self, topic):
        return '/'.join(filter(None, (self.unique_id, topic)))

    def _get_topic_for_entity(self, entity, *, skip_unique_id=False):
        subtopic = entity.get('topic', self.STATE_TOPIC)
        if skip_unique_id:
            return subtopic
        return self._get_topic(subtopic)

    def get_entity_by_name(self, domain: str, name: str):
        return next(
            (e for e in self.entities.get(domain, []) if e['name'] == name),
            None,
        )

    def get_entity_subtopic_from_topic(self, topic: str) -> tuple:
        action_postfix = None
        if topic.startswith(self.unique_id):
            topic = topic[len(self.unique_id):]
        for postfix in [
            self.SET_POSTFIX,
            self.SET_POSITION_POSTFIX,
            self.SET_MODE_POSTFIX,
            self.SET_TARGET_TEMPERATURE_POSTFIX,
        ]:
            if topic.endswith(postfix):
                action_postfix = postfix
                topic = topic[:-len(postfix)]
                break
        return topic.strip('/'), action_postfix

    @property
    def subscribed_topics(self):
        postfix_domains = {
            self.SET_POSTFIX:
                [
                    CLIMATE_DOMAIN, COVER_DOMAIN, LIGHT_DOMAIN, SELECT_DOMAIN,
                    SWITCH_DOMAIN,
                ],
            self.SET_POSITION_POSTFIX: [COVER_DOMAIN],
            self.SET_MODE_POSTFIX: [CLIMATE_DOMAIN],
            self.SET_TARGET_TEMPERATURE_POSTFIX: [CLIMATE_DOMAIN],
        }

        topics = []
        for postfix, domains in postfix_domains.items():
            topics.extend((
                '/'.join(filter(None, (
                    self.unique_id,
                    entity.get('topic', self.STATE_TOPIC),
                    postfix,
                )))
                for cls, items in self.entities.items()
                for entity in items
                if cls in domains
            ))
        return topics

    @property
    def manufacturer(self):
        return self._manufacturer

    @property
    def model(self):
        return self._model

    @property
    def version(self):
        return self._version

    @property
    def dev_id(self):
        return self.mac.replace(':', '').lower()

    @property
    def friendly_id(self):
        # should be used in entity names in homeassistant
        return self.friendly_name or self.dev_id

    @property
    def unique_id(self):
        # name and manufacturer can change while working, e.g. when
        # a device sends his name. To avoid changing topics use
        # the ID based on mac address only
        return f'0x{self.dev_id}'

    @property
    def unique_name(self):
        # can change over time. Don't use it as an identifier
        parts = [self.manufacturer, self.model, self.friendly_id]
        return '_'.join([p.replace(' ', '_') for p in parts if p])

    @property
    def rssi(self):
        return self._rssi

    @rssi.setter
    def rssi(self, value):
        self._rssi = value

    @property
    def linkquality(self):
        if self.rssi is None:
            return None
        return rssi_to_linkquality(self.rssi)

    @property
    def availability_topic(self):
        return self._get_topic('availability')

    @property
    @abc.abstractmethod
    def entities(self) -> ty.Dict[str, ty.Any]:
        return {}

    @abc.abstractmethod
    def get_values_by_entities(self) -> ty.Dict[str, ty.Any]:
        pass

    async def send_availability(self, publish_topic, value: bool):
        await publish_topic(
            topic=self.availability_topic,
            value='online' if value else 'offline',
            nowait=True,
        )

    async def handle_messages(self, *args, **kwargs):
        while True:
            await aio.sleep(1)

    async def update_device_data(self, send_config):
        """
        Call this method on each iteration in handle.
        It will update rssi and config
        """
        if not self.config_sent:
            await send_config()
        if self.client:  # in passive mode, client is None
            props = self.client._properties
            self.rssi = props.get('RSSI')

    def __str__(self):
        return self.unique_name

    def __repr__(self):
        return f'<Device:{str(self)}>'

    async def add_incoming_message(self, topic: str, value):
        await self.message_queue.put({
            'topic': topic,
            'value': value,
        })

    async def get_device_data(self):
        """Here put the initial configuration for the device"""
        pass

    async def get_client(self, **kwargs) -> BleakClient:
        assert self.MAC_TYPE in ('public', 'random')
        return BleakClient(self.mac, address_type=self.MAC_TYPE, **kwargs)

    async def connect(self):
        if self.is_passive:
            return

        self.client = await self.get_client(
            disconnected_callback=self._on_disconnect,
        )
        self.disconnected_event.clear()
        try:
            # 10 is the implicit timeout in bleak client, add 2 more seconds
            # for internal routines
            await aio.wait_for(self.client.connect(), timeout=12)
        except aio.TimeoutError as e:
            self.disconnected_event.set()
            raise ConnectionTimeoutError() from e
        except (Exception, aio.CancelledError):
            self.disconnected_event.set()
            raise
        self._advertisement_seen.clear()
        _LOGGER.info(f'Connected to {self.client.address}')

    def _on_disconnect(self, client, *args):
        _LOGGER.debug(f'Client {client.address} disconnected, device={self}')
        self.disconnected_event.set()

    async def close(self):
        try:
            connected = self.client and self.client.is_connected
        # exception on macos when checking for is_connected()
        except AttributeError:
            connected = True
        if connected:
            await self.client.disconnect()
        await super().close()

    @property
    def entities_with_lqi(self):
        sensor_entities = self.entities.get(SENSOR_DOMAIN, [])
        sensor_entities.append(
            {
                'name': 'linkquality',
                'unit_of_measurement': 'lqi',
                'icon': 'signal',
                'entity_category': 'diagnostic',
                **(
                    {'topic': self.LINKQUALITY_TOPIC}
                    if self.LINKQUALITY_TOPIC else {}
                ),
            },
        )
        return {
            **self.entities,
            SENSOR_DOMAIN: sensor_entities,
        }

    async def _notify_state(self, publish_topic):
        values_by_name = {
            'linkquality': self.linkquality,
            **self.get_values_by_entities(),
        }

        _LOGGER.info(f'[{self}] send state={values_by_name}')

        data_by_topic = defaultdict(dict)
        for domain, entities in self.entities_with_lqi.items():
            for entity in entities:
                name = entity['name']
                if name not in values_by_name:
                    continue

                value = values_by_name[name]
                content_values = (
                    value if isinstance(value, dict) else {name: value}
                )

                for parameter, val in content_values.items():
                    if domain in [SENSOR_DOMAIN, BINARY_SENSOR_DOMAIN]:
                        val = self.transform_value(val)
                    topic = self._get_topic_for_entity(entity)
                    data_by_topic[topic][parameter] = val
        coros = [
            publish_topic(topic=topic, value=json.dumps(values))
            for topic, values in data_by_topic.items()
        ]
        if coros:
            await aio.gather(*coros)
            self.initial_status_sent = True


class Sensor(Device, abc.ABC):
    # a list of state properties that must be not None at least one of them
    # to send data.
    # E.g. only battery updated, but wait for temperature and humidity
    REQUIRED_VALUES: ty.Sequence[str] = ()
    READ_DATA_IN_ACTIVE_LOOP: bool = False

    def __init__(self, mac, *args, loop, **kwargs) -> None:
        super().__init__(mac, *args, loop=loop, **kwargs)
        self._state = None

    @property
    def entities(self):
        raise NotImplementedError()

    async def get_device_data(self):
        name = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(name, (bytes, bytearray)):
            self._model = name.decode().strip('\0')
        version = await self._read_with_timeout(FIRMWARE_VERSION)
        if isinstance(version, (bytes, bytearray)):
            self._version = version.decode().strip('\0')

    def get_entity_map(self):
        state = {}
        for domain, entities in self.entities.items():
            for entity in entities:
                sensor_name = entity['name']
                value = getattr(self._state, sensor_name, None)
                if value is not None:
                    state[sensor_name] = self.transform_value(value)
        if self.REQUIRED_VALUES and not any(
            state.get(x) for x in self.REQUIRED_VALUES
        ):
            return {}
        return state

    async def _notify_state(self, publish_topic):
        _LOGGER.info(f'[{self}] send state={self._state}')
        state = self.get_entity_map()
        if state:
            state['linkquality'] = self.linkquality
            await publish_topic(
                topic=self._get_topic(self.STATE_TOPIC),
                value=json.dumps(state),
            )

    async def do_active_loop(self, publish_topic):
        await self._notify_state(publish_topic)

    async def do_passive_loop(self, publish_topic):
        await self._notify_state(publish_topic)

    async def handle_active(self, publish_topic, send_config, *args, **kwargs):
        while True:
            await self.update_device_data(send_config)
            if not self.READ_DATA_IN_ACTIVE_LOOP and not self._state:
                await aio.sleep(self.NOT_READY_SLEEP_INTERVAL)
                continue

            await self.do_active_loop(publish_topic)
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)

    async def handle_passive(self, publish_topic, send_config, *args, **kwargs):
        while True:
            if not self._state:
                await aio.sleep(self.NOT_READY_SLEEP_INTERVAL)
                continue

            await self.update_device_data(send_config)
            await self.do_passive_loop(publish_topic)
            await aio.sleep(self.PASSIVE_SLEEP_INTERVAL)

    async def handle(self, *args, **kwargs):
        if self.is_passive:
            return await self.handle_passive(*args, **kwargs)
        return await self.handle_active(*args, **kwargs)


class SubscribeAndSetDataMixin:
    DATA_CHAR: uuid.UUID = None  # type: ignore
    SENSOR_CLASS: ty.Any = None  # type: ignore

    def filter_notifications(self, sender):
        return True

    def process_data(self, data: bytearray):
        self._state = self.SENSOR_CLASS.from_data(data)

    def notification_handler(self, sender, data: bytearray):
        _LOGGER.debug("{0} notification: {1}: {2}".format(
            self,
            sender,
            format_binary(data),
        ))
        if self.filter_notifications(sender):
            self.process_data(data)

    async def get_device_data(self):
        if self.DATA_CHAR:
            await self.client.start_notify(
                self.DATA_CHAR,
                self.notification_handler,
            )
        await super().get_device_data()


class CoverMovementType(Enum):
    STOP = 0
    POSITION = 1


class BaseCover(Device, abc.ABC):
    COVER_ENTITY = 'cover'

    # HA notation. We convert value on setting and receiving data
    CLOSED_POSITION = 0
    OPEN_POSITION = 100

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._state = namedtuple(
            'CoverState',
            (
                'position', 'target_position', 'run_state',
            ),
        )(position=0, target_position=0, run_state=CoverRunState.STOPPED)

    @property
    def entities(self):
        return {
            COVER_DOMAIN: [
                {
                    'name': self.COVER_ENTITY,
                    'topic': self.COVER_ENTITY,
                    'device_class': 'shade',
                },
            ],
            SENSOR_DOMAIN: [
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'illuminance',
                    'device_class': 'illuminance',
                    'unit_of_measurement': 'lx',
                },
            ],
        }

    @abc.abstractmethod
    async def _stop(self):
        pass

    @abc.abstractmethod
    async def _set_position(self, value):
        pass

    @abc.abstractmethod
    async def _update_running_state(self):
        """ This method is called as a short update while
        opening the shades"""

    @abc.abstractmethod
    async def _update_full_state(self):
        """ This method is called to refetch all values from the device"""

    async def _do_movement(self, movement_type: CoverMovementType,
                           target_position: ty.Optional[int]):
        if movement_type == CoverMovementType.POSITION and \
                target_position is not None:
            if self.CLOSED_POSITION <= target_position <= self.OPEN_POSITION:
                await self._set_position(target_position)
                if self._state.position > target_position:
                    self._state.target_position = target_position
                    self._state.run_state = CoverRunState.CLOSING
                elif self._state.position < target_position:
                    self._state.target_position = target_position
                    self._state.run_state = CoverRunState.OPENING
                else:
                    self._state.target_position = None
                    if target_position == self.OPEN_POSITION:
                        self._state.run_state = CoverRunState.OPEN
                    elif target_position == self.CLOSED_POSITION:
                        self._state.run_state = CoverRunState.CLOSED
                    else:
                        self._state.run_state = CoverRunState.STOPPED
            else:
                _LOGGER.error(
                    f'[{self}] Incorrect position value: '
                    f'{repr(target_position)}',
                )
        else:
            await self._stop()
            self._state.run_state = CoverRunState.STOPPED

    async def _handle_message(self, message, publish_topic):
        value = message['value']
        entity_topic, action_postfix = self.get_entity_subtopic_from_topic(
            message['topic'],
        )
        if entity_topic == self._get_topic_for_entity(
                self.get_entity_by_name(COVER_DOMAIN, self.COVER_ENTITY),
                skip_unique_id=True,
        ):
            value = self.transform_value(value)
            target_position = None
            if action_postfix == self.SET_POSTFIX:
                _LOGGER.info(
                    f'[{self}] set mode {entity_topic} to "{value}"',
                )
                if value.lower() == 'open':
                    movement_type = CoverMovementType.POSITION
                    target_position = self.OPEN_POSITION
                elif value.lower() == 'close':
                    movement_type = CoverMovementType.POSITION
                    target_position = self.CLOSED_POSITION
                else:
                    movement_type = CoverMovementType.STOP
            elif action_postfix == self.SET_POSITION_POSTFIX:
                movement_type = CoverMovementType.POSITION
                _LOGGER.info(
                    f'[{self}] set position {entity_topic} to "{value}"',
                )
                try:
                    target_position = int(value)
                except ValueError:
                    pass
            else:
                _LOGGER.warning(
                    f'[{self}] unknown action postfix {action_postfix}',
                )
                return False

            while True:
                try:
                    await self._do_movement(movement_type, target_position)
                    await self._notify_state(publish_topic)
                    break
                except ConnectionError as e:
                    _LOGGER.exception(str(e))
                await aio.sleep(5)
            return True

    async def handle_messages(self, publish_topic, *args, **kwargs):
        while True:
            try:
                if not self.client.is_connected:
                    raise ConnectionError()
                message = await aio.wait_for(
                    self.message_queue.get(),
                    timeout=60,
                )
            except aio.TimeoutError:
                await aio.sleep(1)
                continue
            await self._handle_message(message, publish_topic)


class ClimateMode(Enum):
    OFF = 'off'
    HEAT = 'heat'


class BaseClimate(Device, abc.ABC):
    CLIMATE_ENTITY = 'climate'
    MODES: ty.Iterable[ClimateMode] = ()

    @property
    def entities(self):
        return {
            CLIMATE_DOMAIN: [
                {
                    'name': self.CLIMATE_ENTITY,
                    'modes': [x.value for x in self.MODES],
                },
            ],
        }

    @abc.abstractmethod
    async def _set_target_temperature(self, value):
        pass

    @abc.abstractmethod
    async def _switch_mode(self, next_mode):
        pass

    async def _handle_message(self, message, publish_topic):
        value = message['value']
        entity_topic, action_postfix = self.get_entity_subtopic_from_topic(
            message['topic'],
        )
        if entity_topic == self._get_topic_for_entity(
                self.get_entity_by_name(CLIMATE_DOMAIN, self.CLIMATE_ENTITY),
                skip_unique_id=True,
        ):
            if action_postfix == self.SET_MODE_POSTFIX:
                _LOGGER.info(
                    f'[{self}] set mode {entity_topic} to "{value}"',
                )
                try:
                    value = ClimateMode(value.lower())
                except ValueError:
                    _LOGGER.warning(f"{self} Incorrect mode {value}")
                    return False
                else:
                    if value not in self.MODES:
                        _LOGGER.warning(f"{self} Incorrect mode {value}")
                        return False
                state_change_coro = self._switch_mode(value)
            elif action_postfix == self.SET_TARGET_TEMPERATURE_POSTFIX:
                try:
                    target_temperature = float(value)
                except ValueError:
                    _LOGGER.exception("Incorrect temperature")
                    return False
                _LOGGER.info(
                    f'[{self}] set temperature {entity_topic} to "{value}"',
                )
                state_change_coro = \
                    self._set_target_temperature(target_temperature)
            else:
                _LOGGER.warning(
                    f'[{self}] unknown action postfix {action_postfix}',
                )
                return False

            if not state_change_coro:
                return False

            while True:
                try:
                    await state_change_coro
                    await self._notify_state(publish_topic)
                    break
                except ConnectionError as e:
                    _LOGGER.exception(str(e))
                await aio.sleep(5)
            return True

    async def handle_messages(self, publish_topic, *args, **kwargs):
        while True:
            try:
                if not self.client.is_connected:
                    raise ConnectionError()
                message = await aio.wait_for(
                    self.message_queue.get(),
                    timeout=60,
                )
            except aio.TimeoutError:
                await aio.sleep(1)
                continue
            await self._handle_message(message, publish_topic)



import abc
import asyncio as aio
import datetime
import logging
import uuid

from ble2mqtt.devices.base import BaseDevice
from ble2mqtt.protocols.base import (BaseCommand, BLEQueueMixin,
                                     SendAndWaitReplyMixin)
from ble2mqtt.utils import format_binary

_LOGGER = logging.getLogger(__name__)


CMD_SET_TIME = 0xaa
CMD_SET_NOTIFY_PERIOD = 0xae
CMD_READ_VALUE = 0xab
CMD_REQUEST_CALIBRATION = 0xad
CMD_RESET = 0xee


class WP6003Command(BaseCommand):
    pass


class WP6003Protocol(SendAndWaitReplyMixin, BLEQueueMixin, BaseDevice,
                     abc.ABC):
    RX_CHAR: uuid.UUID = None  # type: ignore
    TX_CHAR: uuid.UUID = None  # type: ignore

    async def protocol_start(self):
        assert self.RX_CHAR
        await self.client.start_notify(
            self.RX_CHAR,
            self.notification_callback,
        )

    async def send_reset(self):
        await self.send_command(bytes([CMD_RESET]), True, timeout=3)

    async def write_time(self):
        now = datetime.datetime.utcnow()
        set_time_cmd = bytes([
            CMD_SET_TIME,
            now.year - 2000,
            now.month,
            now.day,
            now.hour,
            now.minute,
            now.second,
        ])
        # consider process response here
        return await self.send_command(set_time_cmd, True, timeout=3)

    async def read_value(self) -> bytes:
        return await self.send_command(
            bytes([CMD_READ_VALUE]),
            wait_reply=True,
            timeout=3,
        )

    async def send_command(self, cmd: bytes = b'',
                           wait_reply=False, timeout=10):
        command = WP6003Command(cmd, wait_reply=wait_reply, timeout=timeout)
        self.clear_ble_queue()
        await self.cmd_queue.put(command)
        return await aio.wait_for(command.answer, timeout)

    async def process_command(self, command: WP6003Command):
        _LOGGER.debug(f'... send cmd {format_binary(command.cmd)}')
        self.clear_ble_queue()
        cmd_resp = await aio.wait_for(
            self.client.write_gatt_char(self.TX_CHAR, command.cmd),
            timeout=command.timeout,
        )
        if not command.wait_reply:
            if command.answer.cancelled():
                return
            command.answer.set_result(cmd_resp)
            return

        ble_notification = await self.ble_get_notification(command.timeout)

        # extract payload from container
        cmd_resp = bytes(ble_notification[1])
        if command.answer.cancelled():
            return
        command.answer.set_result(cmd_resp)



import abc
import logging
import uuid

from ..devices.base import BaseDevice
from ..utils import format_binary
from .base import BLEQueueMixin

_LOGGER = logging.getLogger(__name__)

# command IDs
AM43_CMD_MOVE = 0x0a
AM43_CMD_GET_BATTERY = 0xa2
AM43_CMD_GET_ILLUMINANCE = 0xaa
AM43_CMD_GET_POSITION = 0xa7
AM43_CMD_SET_POSITION = 0x0d
AM43_NOTIFY_POSITION = 0xa1

AM43_RESPONSE_ACK = 0x5a
AM43_RESPONSE_NACK = 0xa5
# 9a a8 0a 01 00 7f 09 1e 01 64 7f 11 00 5a
AM43_REPLY_UNKNOWN1 = 0xa8
# 9a a9 10 00 00 00 11 00 00 00 00 01 00 00 11 00 00 00 00 22
AM43_REPLY_UNKNOWN2 = 0xa9


class AM43Protocol(BLEQueueMixin, BaseDevice, abc.ABC):
    DATA_CHAR: uuid.UUID = None  # type: ignore

    def notification_callback(self, sender_handle: int, data: bytearray):
        self.process_data(data)
        super().notification_callback(sender_handle, data)

    @staticmethod
    def _convert_position(value):
        return 100 - value

    @classmethod
    def convert_to_device(cls, value):
        return cls._convert_position(value)

    @classmethod
    def convert_from_device(cls, value):
        return cls._convert_position(value)

    async def send_command(self, cmd_id, data: list,
                           wait_reply=True, timeout=25):
        _LOGGER.debug(f'[{self}] - send command 0x{cmd_id:x} {data}')
        cmd = bytearray([0x9a, cmd_id, len(data)] + data)
        csum = 0
        for x in cmd:
            csum = csum ^ x
        cmd += bytearray([csum])

        self.clear_ble_queue()
        await self.client.write_gatt_char(self.DATA_CHAR, cmd)
        ret = None
        if wait_reply:
            _LOGGER.debug(f'[{self}] waiting for reply')
            ble_notification = await self.ble_get_notification(timeout)
            _LOGGER.debug(f'[{self}] reply: {repr(ble_notification[1])}')
            ret = bytes(ble_notification[1])
        return ret

    async def _get_position(self):
        await self.send_command(AM43_CMD_GET_POSITION, [0x01], True)

    async def _get_battery(self):
        await self.send_command(AM43_CMD_GET_BATTERY, [0x01], True)

    async def _get_illuminance(self):
        await self.send_command(AM43_CMD_GET_ILLUMINANCE, [0x01], True)

    async def _set_position(self, value):
        await self.send_command(
            AM43_CMD_SET_POSITION,
            [self.convert_to_device(int(value))],
            True,
        )

    async def _stop(self):
        await self.send_command(AM43_CMD_MOVE, [0xcc])

    async def _open(self):
        # not used
        await self.send_command(AM43_CMD_MOVE, [0xdd])

    async def _close(self):
        # not used
        await self.send_command(AM43_CMD_MOVE, [0xee])

    async def _update_full_state(self):
        await self._get_position()
        await self._get_battery()
        await self._get_illuminance()

    @abc.abstractmethod
    def handle_battery(self, value):
        pass

    @abc.abstractmethod
    def handle_position(self, value):
        pass

    @abc.abstractmethod
    def handle_illuminance(self, value):
        pass

    def process_data(self, data: bytearray):
        if data[1] == AM43_CMD_GET_BATTERY:
            # b'\x9a\xa2\x05\x00\x00\x00\x00Ql'
            self.handle_battery(int(data[7]))
        elif data[1] == AM43_NOTIFY_POSITION:
            self.handle_position(self.convert_from_device(int(data[4])))
        elif data[1] == AM43_CMD_GET_POSITION:
            # [9a a7 07 0e 32 00 00 00 00 30 36]
            # Bytes in this packet are:
            #  3: Configuration flags, bits are:
            #    1: direction
            #    2: operation mode
            #    3: top limit set
            #    4: bottom limit set
            #    5: has light sensor
            #  4: Speed setting
            #  5: Current position
            #  6,7: Shade length.
            #  8: Roller diameter.
            #  9: Roller type.

            self.handle_position(self.convert_from_device(int(data[5])))
        elif data[1] == AM43_CMD_GET_ILLUMINANCE:
            # b'\x9a\xaa\x02\x00\x002'
            self.handle_illuminance(int(data[4]) * 12.5)
        elif data[1] in [AM43_CMD_MOVE, AM43_CMD_SET_POSITION]:
            if data[3] != AM43_RESPONSE_ACK:
                _LOGGER.error(f'[{self}] Problem with moving: NACK')
        elif data[1] in [AM43_REPLY_UNKNOWN1, AM43_REPLY_UNKNOWN2]:
            # [9a a8 00 32]
            # [9a a9 10 00 00 00 11 00 00 00 00 01 00 00 11 00 00 00 00 22]
            pass
        else:
            _LOGGER.error(
                f'{self} BLE notification unknown response '
                f'[{format_binary(data)}]',
            )



import abc
import asyncio as aio
import logging
import struct
import uuid

from ..devices.base import BaseDevice
from ..utils import color_rgb_to_rgbw, color_rgbw_to_rgb, format_binary
from .base import BaseCommand, BLEQueueMixin, SendAndWaitReplyMixin

_LOGGER = logging.getLogger(__name__)

CMD_COLOR = 0x35
CMD_BRIGHTNESS = 0x57
CMD_NAME = 0x58


class AveaCommand(BaseCommand):
    pass


class AveaProtocol(BLEQueueMixin, SendAndWaitReplyMixin, BaseDevice, abc.ABC):
    DATA_CHAR: uuid.UUID = None  # type: ignore

    async def get_device_data(self):
        if self.DATA_CHAR:
            await self.client.start_notify(
                self.DATA_CHAR,
                self.notification_callback,
            )

    def notification_callback(self, sender_handle: int, data: bytearray):
        self.process_data(data)
        super().notification_callback(sender_handle, data)

    async def send_command(self, cmd: bytes = b'',
                           wait_reply=False, timeout=10):
        command = AveaCommand(cmd, wait_reply=wait_reply, timeout=timeout)
        await self.cmd_queue.put(command)
        return await aio.wait_for(command.answer, timeout)

    async def process_command(self, command: AveaCommand):
        _LOGGER.debug(f'... send cmd {format_binary(command.cmd)}')
        self.clear_ble_queue()
        cmd_resp = await aio.wait_for(
            self.client.write_gatt_char(self.DATA_CHAR, command.cmd, True),
            timeout=command.timeout,
        )
        if not command.wait_reply:
            if command.answer.cancelled():
                return
            command.answer.set_result(cmd_resp)
            return

        ble_notification = await self.ble_get_notification(command.timeout)

        # extract payload from container
        cmd_resp = bytes(ble_notification[1])
        if command.answer.cancelled():
            return
        command.answer.set_result(cmd_resp)

    @abc.abstractmethod
    def handle_color(self, value):
        """Handle color tuple (r,g,b)"""
        pass

    @abc.abstractmethod
    def handle_brightness(self, value):
        """Handle brightness value"""
        pass

    @abc.abstractmethod
    def handle_name(self, value):
        pass

    def process_data(self, data: bytearray):
        # b'5\x00\x00\x00\x00\x00\xff\x15\x00 \xff>\x00\x00\x00\x16\x00 \x00?'
        if data[0] == CMD_COLOR:
            (
                cur_white, cur_blue, cur_green, cur_red,
                white, blue, green, red,
            ) = struct.unpack('<HHHHHHHH', data[4:])

            rgbw = (
                (red ^ 0x3000),
                (green ^ 0x2000),
                (blue ^ 0x1000),
                white,
            )
            self.handle_color(tuple(x // 16 for x in color_rgbw_to_rgb(*rgbw)))
        elif data[0] == CMD_BRIGHTNESS:
            self.handle_brightness(int.from_bytes(data[1:], 'little') // 16)
        # elif data[0] == CMD_NAME:
        #     self.handle_name(data[1:].decode().strip('\0'))

    async def read_name(self, timeout=3):
        response = await self.send_command(bytes([CMD_NAME]), True, timeout)
        return response[1:].decode(errors='ignore').strip('\0')

    async def read_state(self, timeout=3):
        color_read_cmd = bytes([CMD_COLOR])

        # consider process response here
        await self.send_command(color_read_cmd, True, timeout)
        brightness_read_cmd = bytes([CMD_BRIGHTNESS])
        await self.send_command(brightness_read_cmd, True, timeout)

    @staticmethod
    def convert_value(val):
        # handle PWM if any at max level
        if val <= 0:
            return 0
        elif val < 255:
            return val * 16
        return 4095

    @staticmethod
    def get_color_cmd(w=2000, r=0, g=0, b=0, delay=100):
        """Return the command for the specified colors"""

        fading = delay.to_bytes(2, byteorder='little')
        unknown_magic = bytes([0x0a, 0x00])
        white = (int(w) | int(0x8000)).to_bytes(2, byteorder='little')
        red = (int(r) | int(0x3000)).to_bytes(2, byteorder='little')
        green = (int(g) | int(0x2000)).to_bytes(2, byteorder='little')
        blue = (int(b) | int(0x1000)).to_bytes(2, byteorder='little')

        return (
                bytes([CMD_COLOR]) + fading + unknown_magic +
                white + red + green + blue
        )

    @staticmethod
    def get_brightness_cmd(brightness):
        """Return the command for the specified brightness"""

        return (
                bytes([CMD_BRIGHTNESS]) +
                brightness.to_bytes(2, byteorder='little')
        )

    async def write_color(self, r, g, b):
        _r, _g, _b, _w = color_rgb_to_rgbw(r, g, b)
        cmd = self.get_color_cmd(
            self.convert_value(_w),
            self.convert_value(_r),
            self.convert_value(_g),
            self.convert_value(_b),
        )
        _LOGGER.info(
            f'Writing color ({(_r, _g, _b, _w)}): {format_binary(cmd)}',
        )
        await self.send_command(cmd)

    async def write_brightness(self, brightness):
        cmd = self.get_brightness_cmd(self.convert_value(brightness))
        _LOGGER.info(f'Writing brightness {brightness}: {format_binary(cmd)}')
        await self.send_command(cmd)



import logging
import typing as ty
import uuid
from dataclasses import dataclass

from bleak.backends.device import BLEDevice

from ..protocols.xiaomi import parse_fe95_advert
from ..utils import format_binary
from .base import ConnectionMode
from .uuids import BATTERY
from .xiaomi_base import XiaomiHumidityTemperature

_LOGGER = logging.getLogger(__name__)

MJHT_DATA = uuid.UUID('226caa55-6476-4566-7562-66734470666d')
ADVERTISING = uuid.UUID('0000fe95-0000-1000-8000-00805f9b34fb')


@dataclass
class SensorState:
    battery: int = 0
    temperature: float = 0
    humidity: float = 0
    voltage: ty.Optional[float] = None

    @classmethod
    def from_data(cls, sensor_data, battery_data):
        # b'T=23.6 H=39.6\x00'
        t, h = tuple(
            float(x.split('=')[1])
            for x in sensor_data.decode().strip('\0').split(' ')
        )
        battery = int(ord(battery_data)) if battery_data else 0
        return cls(
            temperature=t,
            humidity=h,
            battery=battery,
        )


class XiaomiHumidityTemperatureV1(XiaomiHumidityTemperature):
    NAME = 'xiaomihtv1'
    DATA_CHAR = MJHT_DATA
    BATTERY_CHAR = BATTERY
    SENSOR_CLASS = SensorState
    SUPPORT_PASSIVE = True
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_POLL_WITH_DISCONNECT

    def filter_notifications(self, sender):
        # sender is 0xd or several requests it becomes
        # /org/bluez/hci0/dev_58_2D_34_32_E0_69/service000c/char000d
        return (
            sender == 0xd or
            isinstance(sender, str) and sender.endswith('000d')
        )

    def handle_advert(self, scanned_device: BLEDevice, adv_data):
        service_data = adv_data.service_data
        adv_data = service_data.get(str(ADVERTISING))

        if adv_data:
            # frctrl devic id <----- mac -----> type len <-- data -->
            # [50 20 aa 01 e4 69 e0 32 34 2d 58 0d 10 04 df 00 55 01]
            # [50 20 aa 01 80 69 e0 32 34 2d 58 0d 10 04 d6 00 29 01]

            try:
                parsed_advert = parse_fe95_advert(bytes(adv_data))
            except (ValueError, IndexError):
                _LOGGER.exception(f'{self} Cannot parse advert packet')
                return

            if self._state is None:
                self._state = self.SENSOR_CLASS()
            for k, v in parsed_advert.items():
                setattr(self._state, k, v)

            _LOGGER.debug(
                f'Advert received for {self}, {format_binary(adv_data)}, '
                f'current state: {self._state}',
            )



import abc
import logging
import typing as ty
import uuid
from enum import Enum

from ble2mqtt.devices.base import BaseDevice

from ..utils import format_binary
from .base import BLEQueueMixin

_LOGGER = logging.getLogger(__name__)


class MotorTriggerCommandCodes(Enum):
    ADD = 0x13
    REMOVE = 0x23
    READ = 0x33
    EDIT = 0x43
    CLEAR_ALL = 0x63


class MotorCommandCodes(Enum):
    STOP = 0x00
    STOP_AT_NEXT_STEP = 0x01
    STOP_AT_NEXT_STEP_STATE = 0x03
    STEP_UP = 0x68
    UP = 0x69
    STEP_DOWN = 0x86
    DOWN = 0x96
    LOW_BATTERY = 0xff


class ConfigCommandCodes(Enum):
    MOTOR_SPEED = 1
    MOTOR_DIRECTION = 2
    MOTOR_SPEED_TRIGGER = 3
    PID = 4
    GEO_POSITION = 5
    LOCAL_TIME_OFFSET = 6
    MOTOR_ACCELERATION = 7
    MOTOR_DECELERATION = 8
    MOTOR_USTALL_ACCELERATION = 9
    INCREASE_ENCODER_BY2 = 10
    INCREASE_ENCODER_BY4 = 11
    BOOT_SEQ = 12
    RESET_REASON = 13
    STOP_REASON = 14
    POF_COUNT = 15
    SLIP_LENGTH = 16
    ENC_MAX = 17
    ENC_CUR = 18
    SLIP_INTERVAL = 19
    POSITION_MOVE_TOTAL = 20
    MOTOR_MOVE_TOTAL = 21
    IN_CALIBRATION_MODE = 22
    SUNRISE_SUNSET = 23
    MOTOR_CURRENT = 24
    QUERY = 255


class SomaProtocol(BLEQueueMixin, BaseDevice, abc.ABC):
    POSITION_CHAR: uuid.UUID
    MOTOR_CHAR: uuid.UUID
    SET_POSITION_CHAR: uuid.UUID
    BATTERY_CHAR: uuid.UUID
    CHARGING_CHAR: uuid.UUID
    CONFIG_CHAR: uuid.UUID

    LIGHT_COEFF_TO_LUX = 10

    def notification_callback(self, sender_handle: int, data: bytearray):
        _LOGGER.debug(
            f'{self} notification: {sender_handle}: {format_binary(data)}')
        if sender_handle == 71:
            # CONFIG_CHAR
            if data[0] == ConfigCommandCodes.QUERY.value:
                return super().notification_callback(sender_handle, data)
            if data[0] == ConfigCommandCodes.MOTOR_SPEED.value:
                # values are 0x0, 0x3, 0x69, 0x96
                self._handle_motor_run_state(MotorCommandCodes(data[2]))
        elif sender_handle == 32:
            # POSITION_CHAR
            self._handle_position(self._convert_position(data[0]))
        elif sender_handle == 66:
            # CHARGING_CHAR
            self._handle_charging(**self._parse_charge_response(data))

    @staticmethod
    def _convert_position(value):
        return 100 - value

    @abc.abstractmethod
    def _handle_position(self, value):
        pass

    @abc.abstractmethod
    def _handle_charging(self, *, charging_level, panel_level):
        pass

    @abc.abstractmethod
    def _handle_motor_run_state(self, run_state: MotorCommandCodes):
        pass

    async def _get_position(self):
        response = await self.client.read_gatt_char(self.POSITION_CHAR)
        _LOGGER.debug(f'{self} _get_position: [{format_binary(response)}]')
        return self._convert_position(response[0])

    async def _get_target_position(self):
        response = await self.client.read_gatt_char(self.SET_POSITION_CHAR)
        _LOGGER.debug(
            f'{self} _get_target_position: [{format_binary(response)}]',
        )
        return self._convert_position(response[0])

    async def _get_battery(self):
        response = await self.client.read_gatt_char(self.BATTERY_CHAR)
        _LOGGER.debug(
            f'{self} _get_battery: [{format_binary(response)}]',
        )
        return int(min(100.0, response[0] / 75 * 100))

    async def _get_light_and_panel(self):
        response = await self.client.read_gatt_char(self.CHARGING_CHAR)
        _LOGGER.debug(
            f'{self} _get_light_and_panel: [{format_binary(response)}]',
        )
        return self._parse_charge_response(response)

    async def _set_position(self, value):
        value = self._convert_position(value)
        _LOGGER.debug(f'{self} _set_position: {value}')
        await self.client.write_gatt_char(
            self.SET_POSITION_CHAR,
            bytes([value]),
            response=False,
        )

    def _parse_config_response(self, data):
        result = {}
        if data[0] != ConfigCommandCodes.QUERY.value:
            return result

        offset = 2
        while offset < len(data):
            cmd = data[offset]
            length = data[offset + 1]
            offset += 2
            value = data[offset:offset + length]
            if length == 1:
                value = value[0]
            result[ConfigCommandCodes(cmd)] = value
            offset += length
        return result

    def _parse_charge_response(self, data):
        return {
            'charging_level': (
                int.from_bytes(data[:2], byteorder='little') *
                self.LIGHT_COEFF_TO_LUX
            ),
            'panel_level': int.from_bytes(data[2:4], byteorder='little'),
        }

    async def _get_motor_speed(self) -> ty.Optional[int]:
        cmd = bytes([
            ConfigCommandCodes.QUERY.value,
            0x01,  # length
            ConfigCommandCodes.MOTOR_SPEED.value,
        ])
        self.clear_ble_queue()
        await self.client.write_gatt_char(
            self.CONFIG_CHAR,
            cmd,
            response=True,
        )
        ble_notification = await self.ble_get_notification(timeout=10)
        parsed = self._parse_config_response(ble_notification[1])
        _LOGGER.debug(f'_motor_speed parsed {parsed}')
        return parsed.get(ConfigCommandCodes.MOTOR_SPEED, None)

    async def _set_motor_speed(self, value):
        value = value & 0xff
        _LOGGER.debug(f'{self} _set_motor_speed: {value}')
        cmd = bytes([ConfigCommandCodes.MOTOR_SPEED.value, 0x01, value])
        _LOGGER.debug(f'{self} _set_motor_speed cmd [{format_binary(cmd)}]')
        await self.client.write_gatt_char(
            self.CONFIG_CHAR,
            cmd,
            response=True,
        )

    async def _stop(self):
        resp = await self.client.write_gatt_char(
            self.MOTOR_CHAR,
            bytes([MotorCommandCodes.STOP.value]),
            response=True,
        )
        _LOGGER.debug(f'{self} _stop: {resp}')

    # async def _open(self):
    #     await self.client.write_gatt_char(
    #         self.MOTOR_CHAR,
    #         bytes([MotorCommandCodes.UP.value]),
    #         response=True,
    #     )
    #
    # async def _close(self):
    #     await self.client.write_gatt_char(
    #         self.MOTOR_CHAR,
    #         bytes([MotorCommandCodes.DOWN.value]),
    #         response=False,
    #     )



import asyncio as aio
import json
import logging
import os

from ble2mqtt.__version__ import VERSION
from ble2mqtt.ble2mqtt import Ble2Mqtt

from .devices import registered_device_types

_LOGGER = logging.getLogger(__name__)

is_shutting_down: aio.Lock = aio.Lock()


async def shutdown(loop, service: Ble2Mqtt, signal=None):
    """Cleanup tasks tied to the service's shutdown."""
    if is_shutting_down.locked():
        return
    async with is_shutting_down:
        if signal:
            _LOGGER.info(f"Received exit signal {signal.name}...")
        _LOGGER.info("Closing ble2mqtt service")
        await service.close()
        tasks = [t for t in aio.all_tasks() if t is not aio.current_task()]

        [task.cancel() for task in tasks]

        _LOGGER.info(f"Cancelling {len(tasks)} outstanding tasks")
        try:
            await aio.wait_for(
                aio.gather(*tasks, return_exceptions=True),
                timeout=10,
            )
        except (Exception, aio.CancelledError):
            _LOGGER.exception(f'Cancelling caused error: {tasks}')
        loop.stop()


def handle_exception(loop, context, service):
    _LOGGER.error(f"Caught exception: {context}")
    loop.default_exception_handler(context)
    exception_str = context.get('task') or context.get('future') or ''
    exception = context.get('exception')
    if 'BleakClientBlueZDBus._disconnect_monitor()' in \
            str(repr(exception_str)):
        # There is some problem when Bleak waits for disconnect event
        # and asyncio destroys the task and raises
        # Task was destroyed but it is pending!
        # Need further investigating.
        # Skip this exception for now.
        _LOGGER.info("Ignore this exception.")
        return

    if "'NoneType' object has no attribute" in \
            str(repr(exception)):
        # lambda _: self._disconnecting_event.set()
        # AttributeError: 'NoneType' object has no attribute 'set'
        # await self._disconnect_monitor_event.wait()
        # AttributeError: 'NoneType' object has no attribute 'wait'
        _LOGGER.info("Ignore this exception.")
        return

    if isinstance(exception, BrokenPipeError):
        # task = asyncio.ensure_future(self._cleanup_all())
        # in bluezdbus/client.py: _parse_msg() can fail while remove_match()
        _LOGGER.info("Ignore this exception.")
        return

    _LOGGER.info("Shutting down...")
    aio.create_task(shutdown(loop, service))


async def amain(config):
    loop = aio.get_running_loop()

    service = Ble2Mqtt(
        reconnection_interval=10,
        loop=loop,
        host=config['mqtt_host'],
        port=config['mqtt_port'],
        user=config.get('mqtt_user'),
        password=config.get('mqtt_password'),
        base_topic=config['base_topic'],
        mqtt_config_prefix=config['mqtt_config_prefix'],
    )

    loop.set_exception_handler(
        lambda *args: handle_exception(*args, service=service),
    )

    devices = config.get('devices') or []
    for device in devices:
        try:
            mac = device.pop('address')
            typ = device.pop('type')
        except (ValueError, IndexError):
            continue
        klass = registered_device_types[typ]
        service.register(
            klass,
            mac=mac,
            loop=loop,
            **device,
        )

    try:
        await service.start()
    except KeyboardInterrupt:
        _LOGGER.info('Exiting...')
    finally:
        await service.close()


def main():
    os.environ.setdefault('BLE2MQTT_CONFIG', '/etc/ble2mqtt.json')
    config = {}
    if os.path.exists(os.environ['BLE2MQTT_CONFIG']):
        try:
            with open(os.environ['BLE2MQTT_CONFIG'], 'r') as f:
                config = json.load(f)
        except FileNotFoundError:
            pass

    config = {
        'mqtt_host': 'localhost',
        'mqtt_port': 1883,
        'base_topic': 'ble2mqtt',
        'mqtt_config_prefix': 'b2m_',
        'log_level': 'INFO',
        # 'hci_device': 'hci0',
        **config,
    }

    logging.basicConfig(level=config['log_level'].upper())
    # logging.getLogger('bleak.backends.bluezdbus.scanner').setLevel('INFO')
    _LOGGER.info(f'Starting BLE2MQTT version {VERSION}')

    try:
        aio.run(amain(config), debug=(config['log_level'].upper() == 'DEBUG'))
    except KeyboardInterrupt:
        pass
    _LOGGER.info('Bye.')


if __name__ == '__main__':
    main()



from typing import Tuple

MAX_RSSI = 0
MIN_RSSI = -100


def format_binary(data: bytes, delimiter=' '):
    return delimiter.join(format(x, '02x') for x in data)


def cr2032_voltage_to_percent(mvolts: int):
    coeff = 0.8  # >2.9V counts as 100% = (2900 - 2100)/100
    return max(min(int(round((mvolts/1000 - 2.1)/coeff, 2) * 100), 100), 0)


def rssi_to_linkquality(rssi):
    return max(int(round(255 * (rssi - MIN_RSSI) / (MAX_RSSI - MIN_RSSI))), 0)


# code from home assisstant

def _match_max_scale(input_colors, output_colors):
    """Match the maximum value of the output to the input."""
    max_in = max(input_colors)
    max_out = max(output_colors)
    if max_out == 0:
        factor = 0.0
    else:
        factor = max_in / max_out
    return tuple(int(round(i * factor)) for i in output_colors)


def color_rgb_to_rgbw(r: int, g: int, b: int) -> Tuple[int, int, int, int]:
    """Convert an rgb color to an rgbw representation."""
    # Calculate the white channel as the minimum of input rgb channels.
    # Subtract the white portion from the remaining rgb channels.
    w = min(r, g, b)
    rgbw = (r - w, g - w, b - w, w)

    # Match the output maximum value to the input. This ensures the full
    # channel range is used.
    return _match_max_scale((r, g, b), rgbw)  # type: ignore


def color_rgbw_to_rgb(r: int, g: int, b: int, w: int) -> Tuple[int, int, int]:
    """Convert an rgbw color to an rgb representation."""
    # Add the white channel to the rgb channels.
    rgb = (r + w, g + w, b + w)

    # Match the output maximum value to the input. This ensures the
    # output doesn't overflow.
    return _match_max_scale((r, g, b, w), rgb)  # type: ignore



import abc
import datetime
import logging
import struct
import uuid
from dataclasses import dataclass
from enum import Enum

from bleak import BleakError

from ble2mqtt.devices.base import BaseDevice

from ..utils import format_binary
from .base import BLEQueueMixin

_LOGGER = logging.getLogger(__name__)


class EnstoNotAuthorized(Exception):
    pass


class ActiveMode(Enum):
    MANUAL = 1
    CALENDAR = 2
    VACATION = 3


@dataclass
class Measurements:
    target_temperature: float
    temperature: float
    relay_is_on: bool
    alarm_code: int
    active_mode: ActiveMode
    boost_is_on: bool
    potentiometer: int
    boost_minutes: int = 0
    boost_minutes_left: int = 0


class EnstoProtocol(BLEQueueMixin, BaseDevice, abc.ABC):
    MEASUREMENTS_CHAR: uuid.UUID = None  # type: ignore
    VACATION_CHAR: uuid.UUID = None  # type: ignore
    DATE_CHAR: uuid.UUID = None  # type: ignore
    CUSTOM_MEMORY_SLOT_CHAR: uuid.UUID = None  # type: ignore
    AUTH_CHAR: uuid.UUID = None  # type: ignore

    def __init__(self, *args, **kwargs):
        self._reset_id = None
        super().__init__(*args, **kwargs)
        self._heater_potentiometer_temperature = 20.0  # temp on potentiometer

    async def protocol_start(self):
        await self.auth()
        await self.set_date()

    async def auth(self):
        # Need to check if key is provided. Otherwise, read the reset_ket from
        # characteristic
        if not self._reset_id:
            # TODO: pairing in python code doesn't work. Use one-time bluetoothctl pairing
            # _LOGGER.info('pairing...')
            # await self.client.pair()
            _LOGGER.info(f'{self} reading RESET_ID_CHAR {self.AUTH_CHAR}')
            try:
                data = await self.client.read_gatt_char(self.AUTH_CHAR)
            except BleakError as e:
                if 'NotAuthorized' in str(e):
                    raise EnstoNotAuthorized(
                        f'{self} {self.AUTH_CHAR} is not readable.'
                        f' Switch the thermostat in pairing mode',
                    ) from None
                raise
            _LOGGER.debug(f'{self} reset id: {format_binary(data)}')
            if len(data) >= 10:
                self._reset_id = data[:4]
                _LOGGER.warning(
                    f'{self} [!] Write key {self._reset_id.hex()} '
                    f'to config file for later connection',
                )
            else:
                _LOGGER.error(
                    f'{self} Key is unknown and device is not in pairing mode',
                )
        try:
            await self.client.write_gatt_char(
                self.AUTH_CHAR, self._reset_id, response=True,
            )
        except BleakError as e:
            if 'NotAuthorized' in str(e):
                raise EnstoNotAuthorized(
                    f'{self} has incorrect key: {self._reset_id.hex()}',
                ) from None
            raise

    @staticmethod
    def _parse_measurements(data: bytearray) -> Measurements:
        # first part of reporting data
        target_temperature = \
            int.from_bytes(data[1:3], byteorder='little') / 10
        temperature = \
            int.from_bytes(data[4:6], byteorder='little') / 10
        relay_is_on = data[8] == 1
        alarm_code = int.from_bytes(data[9:13], byteorder='little')

        # 1 - manual, 2 - calendar, 3 - vacation
        active_mode = ActiveMode(data[13])
        boost_is_on = data[15] == 1
        boost_minutes = int.from_bytes(data[16:18], byteorder='little')
        boost_minutes_left = int.from_bytes(data[18:20], byteorder='little')
        potentiometer = data[20]
        return Measurements(
            target_temperature=target_temperature,
            temperature=temperature,
            relay_is_on=relay_is_on,
            alarm_code=alarm_code,
            boost_is_on=boost_is_on,
            boost_minutes=boost_minutes,
            boost_minutes_left=boost_minutes_left,
            potentiometer=potentiometer,
            active_mode=active_mode,
        )

    async def read_measurements(self) -> Measurements:
        # f5 32 00 00 b8 00 00 00 00 00 00 00 00 01 02 00 3c 00 3c
        # 00 00 00 05 00 ff ff ff ff ff ff ff ff ff ff ff ff ff ff

        # d2 34 00 00 d8 00 00 00 00 00 00 00 00 03 02 00 3c 00 3c
        # 00 00 00 05 00 ff ff ff ff ff ff ff ff ff ff ff ff ff ff
        data = await self.client.read_gatt_char(self.MEASUREMENTS_CHAR)
        _LOGGER.debug(f'{self} read_measurements: {format_binary(data)}')
        return self._parse_measurements(data)

    async def set_date(self, tzoffset=3):
        _LOGGER.debug(f'{self} set date to current')
        now = datetime.datetime.utcnow() + datetime.timedelta(hours=tzoffset)
        await self.client.write_gatt_char(
            self.DATE_CHAR,
            struct.pack(
                '<H5B',
                now.year,
                now.month,
                now.day,
                now.hour,
                now.minute,
                now.second,
            ),
        )

    async def set_vacation_mode(self, temperature, enable=True):
        offset_temp = int(
            (temperature - self._heater_potentiometer_temperature) * 100,
        )
        data = struct.pack(
            '<10BhbBB',
            0,  # from year
            1,  # from mon
            1,  # from day
            0,
            0,
            255,  # to year
            12,  # to mon
            31,  # to day
            0,
            0,
            offset_temp,
            0,  # offset percentage
            1 if enable else 0,  # enable
            1 if enable else 0,  # vacation mode
        )
        _LOGGER.debug(f'{self} set vacation mode offset: {offset_temp}, '
                      f'{format_binary(data)}')
        await self.client.write_gatt_char(self.VACATION_CHAR, data)

    async def read_vacation_mode(self) -> bytes:
        # 10 07 01 00 00 10 08 01 00 00 0c fe ec 00 00
        data = await self.client.read_gatt_char(self.VACATION_CHAR)
        return data

    async def read_target_temp(self):
        return int.from_bytes((
            await self.client.read_gatt_char(self.CUSTOM_MEMORY_SLOT_CHAR)
        )[:2], byteorder='little')/10

    async def save_target_temp(self, value: float):
        # to keep it working between app restart we store target temp
        # in an unused characteristic
        await self.client.write_gatt_char(
            self.CUSTOM_MEMORY_SLOT_CHAR,
            int(value*10).to_bytes(2, byteorder='little'),
        )



import abc
import asyncio as aio
import logging
import typing as ty

from ..devices.base import BaseDevice
from ..utils import format_binary

_LOGGER = logging.getLogger(__name__)


class BLEQueueMixin(BaseDevice, abc.ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ble_queue = aio.Queue(loop=self._loop)

    def notification_callback(self, sender_handle: int, data: bytearray):
        """
        This method must be used as notification callback for BLE connection
        """
        _LOGGER.debug(f'Notification: {sender_handle}: {format_binary(data)}')
        self._loop.call_soon_threadsafe(
            self._ble_queue.put_nowait, (sender_handle, data),
        )

    def clear_ble_queue(self):
        if hasattr(self._ble_queue, '_queue'):
            self._ble_queue._queue.clear()

    async def ble_get_notification(self, timeout) -> ty.Tuple[int, bytes]:
        ble_response_task = aio.create_task(self._ble_queue.get())
        disconnect_wait_task = aio.create_task(self.disconnected_event.wait())
        await aio.wait(
            [ble_response_task, disconnect_wait_task],
            timeout=timeout,
            return_when=aio.FIRST_COMPLETED,
        )
        if ble_response_task.done():
            disconnect_wait_task.cancel()
            try:
                await disconnect_wait_task
            except aio.CancelledError:
                pass
            return await ble_response_task
        else:
            ble_response_task.cancel()
            try:
                await ble_response_task
            except aio.CancelledError:
                pass
            raise ConnectionError(
                f'{self} cannot fetch response, device is offline',
            )


class BaseCommand:
    def __init__(self, cmd, *args, wait_reply, timeout, **kwargs):
        self.cmd = cmd
        self.answer = aio.Future()
        self.wait_reply = wait_reply
        self.timeout = timeout


class SendAndWaitReplyMixin(BaseDevice, abc.ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cmd_queue: aio.Queue[BaseCommand] = aio.Queue(loop=self._loop)
        self._cmd_queue_task = aio.ensure_future(
            self._handle_cmd_queue(),
            loop=self._loop,
        )
        self._cmd_queue_task.add_done_callback(
            self._queue_handler_done_callback,
        )

    def clear_cmd_queue(self):
        if hasattr(self.cmd_queue, '_queue'):
            self.cmd_queue._queue.clear()

    async def _handle_cmd_queue(self):
        while True:
            command = await self.cmd_queue.get()
            try:
                await self.process_command(command)
            except aio.CancelledError:
                _LOGGER.exception(f'{self} _handle_cmd_queue is cancelled!')
                raise
            except Exception as e:
                if command and not command.answer.done():
                    command.answer.set_exception(e)
                _LOGGER.exception(
                    f'{self} raise an error in handle_queue, ignore it',
                )

    async def process_command(self, command):
        raise NotImplementedError()

    def _queue_handler_done_callback(self, future: aio.Future):
        exc_info = None
        try:
            exc_info = future.exception()
        except aio.CancelledError:
            pass

        if exc_info is not None:
            exc_info = (  # type: ignore
                type(exc_info),
                exc_info,
                exc_info.__traceback__,
            )
            _LOGGER.exception(
                f'{self} _handle_cmd_queue() stopped unexpectedly',
                exc_info=exc_info,
            )



import asyncio as aio
import logging
import struct
import uuid
from dataclasses import dataclass

from ble2mqtt.devices.base import SENSOR_DOMAIN, ConnectionMode, Sensor
from ble2mqtt.devices.uuids import DEVICE_NAME
from ble2mqtt.protocols.wp6003 import WP6003Protocol

_LOGGER = logging.getLogger(__name__)

TX_CHAR = uuid.UUID('0000fff1-0000-1000-8000-00805f9b34fb')
RX_CHAR = uuid.UUID('0000fff4-0000-1000-8000-00805f9b34fb')


@dataclass
class SensorState:
    FORMAT = '>6B6H'

    temperature: float = 0.0
    tvoc: float = 0.0
    hcho: float = 0.0
    co2: int = 0

    @classmethod
    def from_bytes(cls, response):
        # 0a 15 02 0e 09 1e 00d4 0800 0007 0000 0100 0230
        (
            header,  # 0
            year,  # 1
            month,  # 2
            day,  # 3
            hour,  # 4
            minute,  # 5
            temp,  # 6-7
            _,  # 8-9
            tvoc,  # 10-11
            hcho,  # 12-13
            _,  # 14-15
            co2,  # 16-17
        ) = struct.unpack(cls.FORMAT, response)
        if header != 0x0a:
            raise ValueError('Bad response')
        if tvoc == hcho == 0x3fff:
            raise ValueError('Bad value')
        return cls(
            temperature=temp/10,
            tvoc=tvoc/1000,
            hcho=hcho/1000,
            co2=co2,
        )


class VsonWP6003(WP6003Protocol, Sensor):
    NAME = 'wp6003'
    RX_CHAR = RX_CHAR
    TX_CHAR = TX_CHAR
    ACTIVE_SLEEP_INTERVAL = 20
    MANUFACTURER = 'Vson'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION
    READ_DATA_IN_ACTIVE_LOOP = True

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'tvoc',
                    'unit_of_measurement': 'mg/m³',
                    'icon': 'air-filter',
                },
                {
                    'name': 'hcho',
                    'unit_of_measurement': 'mg/m³',
                    'icon': 'air-filter',
                },
                {
                    'name': 'co2',
                    'unit_of_measurement': 'ppm',
                    'icon': 'molecule-co2',
                },
            ],
        }

    async def read_state(self):
        response = await self.read_value()
        for _ in range(5):
            try:
                self._state = SensorState.from_bytes(response)
            except ValueError as e:
                _LOGGER.warning(f'{self} {repr(e)}')
                await aio.sleep(1)
            else:
                break

    async def get_device_data(self):
        # ignore reading firmware version, it doesn't support it
        name = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(name, (bytes, bytearray)):
            self._model = name.decode().strip('\0')
        await self.protocol_start()
        await self.send_reset()
        await self.write_time()

    async def do_active_loop(self, publish_topic):
        try:
            await aio.wait_for(self.read_state(), 5)
            await self._notify_state(publish_topic)
        except (aio.CancelledError, KeyboardInterrupt):
            raise
        except Exception:
            _LOGGER.exception(f'{self} problem with reading values')



import asyncio as aio
import logging
import uuid

from ..protocols.redmond import (ColorTarget, KettleG200Mode, KettleG200State,
                                 KettleRunState, RedmondKettle200Protocol)
from .base import (LIGHT_DOMAIN, SENSOR_DOMAIN, SWITCH_DOMAIN, ConnectionMode,
                   Device)
from .uuids import DEVICE_NAME

_LOGGER = logging.getLogger(__name__)

UUID_NORDIC_TX = uuid.UUID("6e400002-b5a3-f393-e0a9-e50e24dcca9e")
UUID_NORDIC_RX = uuid.UUID("6e400003-b5a3-f393-e0a9-e50e24dcca9e")

BOIL_ENTITY = 'boil'
HEAT_ENTITY = 'heat'  # not implemented yet
TEMPERATURE_ENTITY = 'temperature'
ENERGY_ENTITY = 'energy'
LIGHT_ENTITY = 'backlight'
STATISTICS_ENTITY = 'statistics'


class RedmondKettle(RedmondKettle200Protocol, Device):
    MAC_TYPE = 'random'
    NAME = 'redmond_rk_g200'
    TX_CHAR = UUID_NORDIC_TX
    RX_CHAR = UUID_NORDIC_RX
    ACTIVE_SLEEP_INTERVAL = 1
    RECONNECTION_SLEEP_INTERVAL = 30
    MANUFACTURER = 'Redmond'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    SEND_DATA_PERIOD = 5  # seconds when boiling
    STANDBY_SEND_DATA_PERIOD_MULTIPLIER = 12  # 12 * 5 seconds in standby mode

    def __init__(self, mac, key='ffffffffffffffff',
                 *args, **kwargs):
        super().__init__(mac, *args, **kwargs)
        assert isinstance(key, str) and len(key) == 16
        self._key = bytes.fromhex(key)
        self._state = None
        self._color = (255, 255, 255)
        self._brightness = 255
        self._statistics = {}
        self._energy = None

        self._send_data_period_multiplier = \
            self.STANDBY_SEND_DATA_PERIOD_MULTIPLIER
        self.initial_status_sent = False

    @property
    def entities(self):
        return {
            SWITCH_DOMAIN: [
                {
                    'name': BOIL_ENTITY,
                    'topic': BOIL_ENTITY,
                    'icon': 'kettle',
                },
            ],
            SENSOR_DOMAIN: [
                {
                    'name': TEMPERATURE_ENTITY,
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': ENERGY_ENTITY,
                    'device_class': 'energy',
                    'unit_of_measurement': 'kWh',
                    'state_class': 'total_increasing',
                    'entity_category': 'diagnostic',
                },
                {
                    'name': STATISTICS_ENTITY,
                    'topic': STATISTICS_ENTITY,
                    'icon': 'chart-bar',
                    'json': True,
                    'main_value': 'number_of_starts',
                    'unit_of_measurement': ' ',
                    'entity_category': 'diagnostic',
                },
            ],
            LIGHT_DOMAIN: [
                {
                    'name': LIGHT_ENTITY,
                    'topic': LIGHT_ENTITY,
                },
            ],
        }

    def get_values_by_entities(self):
        return {
            TEMPERATURE_ENTITY: self._state.temperature,
            ENERGY_ENTITY: self._energy,
            STATISTICS_ENTITY: self._statistics,
            LIGHT_ENTITY: {
                'state': (
                    KettleRunState.ON.name
                    if self._state.state == KettleRunState.ON and
                    self._state.mode == KettleG200Mode.LIGHT
                    else KettleRunState.OFF.name
                ),
                'brightness': 255,
                'color': {
                    'r': self._color[0],
                    'g': self._color[1],
                    'b': self._color[2],
                },
                'color_mode': 'rgb',
            },
        }

    async def get_device_data(self):
        await self.protocol_start()
        await self.login(self._key)
        model = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(model, (bytes, bytearray)):
            self._model = model.decode()
        else:
            # macos can't access characteristic
            self._model = 'G200S'
        version = await self.get_version()
        if version:
            self._version = f'{version[0]}.{version[1]}'
        state = await self.get_mode()
        if state:
            self._state = state
            self.update_multiplier()
            self.initial_status_sent = False
        await self.set_time()
        await self._update_statistics()

    def update_multiplier(self, state: KettleG200State = None):
        if state is None:
            state = self._state
        self._send_data_period_multiplier = (
            1
            if state.state == KettleRunState.ON and state.mode in [
                KettleG200Mode.BOIL,
                KettleG200Mode.HEAT,
            ]
            else self.STANDBY_SEND_DATA_PERIOD_MULTIPLIER
        )

    async def notify_run_state(self, new_state: KettleG200State, publish_topic):
        if not self.initial_status_sent or \
                new_state.state != self._state.state or \
                new_state.mode != self._state.mode:
            state_to_str = {
                True: KettleRunState.ON.name,
                False: KettleRunState.OFF.name,
            }
            boil_mode = state_to_str[
                new_state.mode == KettleG200Mode.BOIL and
                new_state.state == KettleRunState.ON
                ]
            heat_mode = state_to_str[
                new_state.mode == KettleG200Mode.HEAT and
                new_state.state == KettleRunState.ON
                ]
            topics = {
                BOIL_ENTITY: boil_mode,
                HEAT_ENTITY: heat_mode,
            }
            await aio.gather(
                *[
                    publish_topic(topic=self._get_topic(topic), value=value)
                    for topic, value in topics.items()
                ],
                self._notify_state(publish_topic),
            )
            self.initial_status_sent = True
            if self._state != new_state:
                self._state = new_state
                await self._notify_state(publish_topic)
            else:
                self._state = new_state
        else:
            self._state = new_state
        self.update_multiplier()

    async def _update_statistics(self):
        statistics = await self.get_statistics()
        self._statistics = {
            'number_of_starts': statistics['starts'],
            'Energy spent (kWh)': round(statistics['watts_hours']/1000, 2),
            'Working time (minutes)': round(statistics['seconds_run']/60, 1),
        }
        self._energy = round(statistics['watts_hours']/1000, 2)

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        counter = 0
        while True:
            await self.update_device_data(send_config)
            # if boiling notify every 5 seconds, 60 sec otherwise
            new_state = await self.get_mode()
            await self.notify_run_state(new_state, publish_topic)
            counter += 1

            if counter > (
                    self.SEND_DATA_PERIOD * self._send_data_period_multiplier
            ):
                await self._update_statistics()
                await self._notify_state(publish_topic)
                counter = 0
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)

    async def _switch_mode(self, mode, value):
        if value == KettleRunState.ON.name:
            try:
                if self._state.mode != mode:
                    await self.stop()
                await self.set_mode(KettleG200State(mode=mode))
            except ValueError:
                # if the MODE is the same then it returns
                # en error. Treat it as normal
                pass
            await self.run()
            next_state = KettleRunState.ON
        else:
            await self.stop()
            next_state = KettleRunState.OFF
        self.update_multiplier(KettleG200State(state=next_state))

    async def _switch_boil(self, value):
        await self._switch_mode(KettleG200Mode.BOIL, value)

    async def _switch_backlight(self, value):
        await self._switch_mode(KettleG200Mode.LIGHT, value)

    async def handle_messages(self, publish_topic, *args, **kwargs):
        while True:
            try:
                if not self.client.is_connected:
                    raise ConnectionError()
                message = await aio.wait_for(
                    self.message_queue.get(),
                    timeout=60,
                )
            except aio.TimeoutError:
                await aio.sleep(1)
                continue
            value = message['value']
            entity_topic, action_postfix = self.get_entity_subtopic_from_topic(
                message['topic'],
            )
            entity = self.get_entity_by_name(SWITCH_DOMAIN, BOIL_ENTITY)
            if entity_topic == self._get_topic_for_entity(
                entity,
                skip_unique_id=True,
            ):
                value = self.transform_value(value)
                _LOGGER.info(
                    f'[{self}] switch kettle {BOIL_ENTITY} value={value}',
                )
                while True:
                    try:
                        await self._switch_boil(value)
                        # update state to real values
                        await self.get_mode()
                        await aio.gather(
                            publish_topic(
                                topic=self._get_topic_for_entity(entity),
                                value=self.transform_value(value),
                            ),
                            self._notify_state(publish_topic),
                            loop=self._loop,
                        )
                        break
                    except ConnectionError as e:
                        _LOGGER.exception(str(e))
                    await aio.sleep(5)
                continue

            entity = self.get_entity_by_name(LIGHT_DOMAIN, LIGHT_ENTITY)
            if entity_topic == self._get_topic_for_entity(
                entity,
                skip_unique_id=True,
            ):
                _LOGGER.info(f'set backlight {value}')
                if value.get('state'):
                    await self._switch_backlight(value['state'])
                if value.get('color') or value.get('brightness'):
                    if value.get('color'):
                        color = value['color']
                        try:
                            self._color = color['r'], color['g'], color['b']
                        except ValueError:
                            return
                    if value.get('brightness'):
                        self._brightness = value['brightness']
                    await aio.gather(
                        self.set_color(
                            ColorTarget.LIGHT,
                            *self._color,
                            self._brightness,
                        ),
                        self._notify_state(publish_topic),
                        loop=self._loop,
                    )



import abc
import logging
import typing as ty

from ..protocols.xiaomi import XiaomiPoller
from .base import SENSOR_DOMAIN

_LOGGER = logging.getLogger(__name__)


class XiaomiHumidityTemperature(XiaomiPoller, abc.ABC):
    SENSOR_CLASS: ty.Any = None
    # send data only if temperature or humidity is set
    REQUIRED_VALUES = ('temperature', 'humidity')

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'humidity',
                    'device_class': 'humidity',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
            ],
        }

    async def read_and_send_data(self, publish_topic):
        battery = await self._read_with_timeout(self.BATTERY_CHAR)
        data_bytes = await self._stack.get()
        # clear queue
        while not self._stack.empty():
            self._stack.get_nowait()
        self._state = self.SENSOR_CLASS.from_data(data_bytes, battery)
        await self._notify_state(publish_topic)



import abc
import asyncio as aio
import logging
import struct
import time
import uuid
from dataclasses import dataclass
from enum import Enum, IntEnum

from ..devices.base import BaseDevice
from ..utils import format_binary
from .base import BaseCommand, BLEQueueMixin, SendAndWaitReplyMixin

_LOGGER = logging.getLogger(__name__)


BOIL_TIME_RELATIVE_DEFAULT = 0x80


class RedmondError(ValueError):
    pass


class Command(IntEnum):
    VERSION = 0x01
    RUN_CURRENT_MODE = 0x03
    STOP_CURRENT_MODE = 0x04
    WRITE_MODE = 0x05
    READ_MODE = 0x06
    WRITE_TEMPERATURE = 0x0b
    WRITE_DELAY = 0x0c
    WRITE_IONIZATION = 0x1b
    WRITE_COLOR = 0x32
    READ_COLOR = 0x33
    SET_BACKLIGHT_MODE = 0x37
    SET_SOUND = 0x3c
    SET_LOCK = 0x3e
    GET_STATISTICS = 0x47
    GET_STARTS_COUNT = 0x50
    SET_TIME = 0x6e
    AUTH = 0xFF


class KettleG200Mode(IntEnum):
    BOIL = 0x00
    HEAT = 0x01
    LIGHT = 0x03
    UNKNOWN1 = 0x04
    UNKNOWN2 = 0x05
    UNKNOWN3 = 0x06


class CookerAfterCookMode(IntEnum):
    HEAT_AFTER_COOK = 0x00
    OFF_AFTER_COOK = 0x01


class KettleRunState(IntEnum):
    OFF = 0x00
    SETUP_PROGRAM = 0x01  # for cooker
    ON = 0x02  # cooker - delayed start
    HEAT = 0x03  # for cooker
    COOKING = 0x05  # for cooker
    WARM_UP = 0x06  # for cooker


class CookerRunState(IntEnum):
    OFF = 0x00
    SETUP_PROGRAM = 0x01  # for cooker
    DELAYED_START = 0x02  # cooker - delayed start
    HEAT = 0x03  # for cooker
    COOKING = 0x05  # for cooker
    WARM_UP = 0x06  # for cooker


class ColorTarget(Enum):
    BOIL = 0x00
    LIGHT = 0x01


class CookerM200Program(Enum):
    FRYING = 0x0
    RICE = 0x1
    MANUAL = 0x2
    PILAF = 0x3
    STEAM = 0x4
    BAKING = 0x5
    STEWING = 0x6
    SOUP = 0x7
    PORRIDGE = 0x8
    YOGHURT = 0x9
    EXPRESS = 0xa


class CookerSubProgram(Enum):
    NONE = 0
    VEGETABLES = 1
    FISH = 2
    MEAT = 3


@dataclass
class KettleG200State:
    temperature: int = 0
    color_change_period: int = 0xf
    mode: KettleG200Mode = KettleG200Mode.BOIL
    target_temperature: int = 0
    sound: bool = True
    is_blocked: bool = False
    state: KettleRunState = KettleRunState.OFF
    boil_time: int = 0
    error: int = 0

    FORMAT = '<6BH2BH4B'

    @classmethod
    def from_bytes(cls, response):
        # 00 00 00 00 01 16 0f 00 00 00 00 00 00 80 00 00 - wait
        # 00 00 00 00 01 14 0f 00 02 00 00 00 00 80 00 00 - boil
        # 01 00 28 00 01 19 0f 00 00 00 00 00 00 80 00 00 - 40º keep
        (
            mode,  # 0
            submode,  # 1
            target_temp,  # 2
            is_blocked,  # 3
            sound,  # 4
            current_temp,  # 5
            color_change_period,  # 6-7
            state,  # 8
            _,  # 9
            ionization,  # 10,11  # for air purifier
            _,   # 12,
            boil_time_relative,  # 13
            _,  # 14
            error,  # 15
        ) = struct.unpack(cls.FORMAT, response)
        return cls(
            mode=KettleG200Mode(mode),
            target_temperature=target_temp,
            sound=sound,
            temperature=current_temp,
            state=KettleRunState(state),
            boil_time=boil_time_relative - BOIL_TIME_RELATIVE_DEFAULT,
            color_change_period=color_change_period,
            error=error,
        )

    def to_bytes(self):
        return struct.pack(
            self.FORMAT,
            self.mode.value,
            0,
            self.target_temperature,
            1 if self.is_blocked else 0,
            self.sound,
            self.temperature,
            self.color_change_period,
            self.state.value,
            0,
            0,
            0,
            self.boil_time + BOIL_TIME_RELATIVE_DEFAULT,
            0,
            0,  # don't send error
        )


@dataclass
class CookerState:
    program: CookerM200Program = CookerM200Program.RICE
    subprogram: CookerSubProgram = CookerSubProgram.NONE
    target_temperature: int = 0
    program_minutes: int = 0
    timer_minutes: int = 0
    after_cooking_mode: CookerAfterCookMode = \
        CookerAfterCookMode.HEAT_AFTER_COOK
    state: CookerRunState = CookerRunState.OFF
    sound: bool = True
    locked: bool = False

    SET_FORMAT = '<8B'
    FORMAT = f'{SET_FORMAT}6BH'

    @classmethod
    def from_bytes(cls, response):
        # 00 00 00 00 00 00 00 00 00 00 01 00 00 00 00 00 - wait with sound
        # 00 00 64 00 00 00 00 00 00 01 01 00 00 00 00 00 - locked
        # 00 00 96 00 0f 00 0f 01 00 00 00 00 00 00 00 00 - 150º 15 minutes

        (
            program,  # 0
            subprogram,  # 1
            target_temp,  # 2
            program_hours,  # 3
            program_minutes,  # 4
            timer_hours,  # 5
            timer_minutes,  # 6
            after_cooking_mode,  # 7
            state,  # 8
            locked,  # 9
            sound,  # 10
            _,  # 11
            _,   # 12,
            _,  # 13
            error,  # 14,15
        ) = struct.unpack(cls.FORMAT, response)
        return cls(
            program=CookerM200Program(program),
            subprogram=CookerSubProgram(subprogram),
            target_temperature=target_temp,
            after_cooking_mode=CookerAfterCookMode(after_cooking_mode),
            state=CookerRunState(state),
            program_minutes=(program_minutes + program_hours * 60),
            timer_minutes=(timer_minutes + timer_hours * 60),
            sound=bool(sound),
            locked=bool(locked),
        )

    def to_bytes(self):
        return struct.pack(
            self.SET_FORMAT,
            self.program.value,
            self.subprogram.value,
            self.target_temperature,
            self.program_minutes // 60,
            self.program_minutes % 60,
            self.timer_minutes // 60,
            self.timer_minutes % 60,
            self.after_cooking_mode.value,
        )


_COOKER_M200_PREDEFINED_PROGRAMS_VALUES = [
    # program, subprogram, temperature, hours, minutes, dhours, dminutes, heat
    [0x00, 0x00, 0x96, 0x00, 0x0f, 0x00, 0x00, 0x01],
    [0x01, 0x00, 0x64, 0x00, 0x19, 0x00, 0x00, 0x01],
    [0x02, 0x00, 0x64, 0x00, 0x1e, 0x00, 0x00, 0x01],
    [0x03, 0x00, 0x6e, 0x01, 0x00, 0x00, 0x00, 0x01],
    [0x04, 0x00, 0x64, 0x00, 0x19, 0x00, 0x00, 0x01],
    [0x05, 0x00, 0x8c, 0x01, 0x00, 0x00, 0x00, 0x01],
    [0x06, 0x00, 0x64, 0x01, 0x00, 0x00, 0x00, 0x01],
    [0x07, 0x00, 0x64, 0x01, 0x00, 0x00, 0x00, 0x01],
    [0x08, 0x00, 0x64, 0x00, 0x1e, 0x00, 0x00, 0x01],
    [0x09, 0x00, 0x28, 0x08, 0x00, 0x00, 0x00, 0x00],
    [0x0a, 0x00, 0x64, 0x00, 0x1e, 0x00, 0x00, 0x00],
]


COOKER_PREDEFINED_PROGRAMS = {
    CookerM200Program(v[0]).name.lower(): CookerState(
        program=CookerM200Program(v[0]),
        subprogram=CookerSubProgram(v[1]),
        target_temperature=v[2],
        program_minutes=v[3] * 60 + v[4],
        timer_minutes=v[5] * 60 + v[6],
        after_cooking_mode=CookerAfterCookMode(v[7]),
    )
    for v in _COOKER_M200_PREDEFINED_PROGRAMS_VALUES
}


class RedmondCommand(BaseCommand):
    def __init__(self, cmd, payload, *args, **kwargs):
        super().__init__(cmd, *args, **kwargs)
        self.payload = payload


class RedmondBaseProtocol(SendAndWaitReplyMixin, BLEQueueMixin, BaseDevice,
                          abc.ABC):
    MAGIC_START = 0x55
    MAGIC_END = 0xaa

    RX_CHAR: uuid.UUID = None  # type: ignore
    TX_CHAR: uuid.UUID = None  # type: ignore

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._cmd_counter = 0

    def _get_command(self, cmd: int, payload: bytes):
        container = struct.pack(
            '<4B',
            self.MAGIC_START,
            self._cmd_counter,
            cmd,
            self.MAGIC_END,
        )
        self._cmd_counter += 1
        if self._cmd_counter > 100:
            self._cmd_counter = 0
        return bytearray(b'%b%b%b' % (container[:3], payload, container[3:]))

    async def send_command(self, cmd: Command, payload: bytes = b'',
                           wait_reply=True, timeout=25):
        command = RedmondCommand(
            cmd,
            payload=payload,
            wait_reply=wait_reply,
            timeout=timeout,
        )
        await self.cmd_queue.put(command)
        return await aio.wait_for(command.answer, timeout)

    async def process_command(self, command: RedmondCommand):
        cmd = self._get_command(command.cmd.value, command.payload)
        _LOGGER.debug(
            f'... send cmd {command.cmd.value:04x} ['
            f'{format_binary(command.payload, delimiter="")}] '
            f'{format_binary(cmd)}',
        )
        self.clear_ble_queue()
        cmd_resp = await aio.wait_for(
            self.client.write_gatt_char(self.TX_CHAR, cmd, True),
            timeout=command.timeout,
        )
        if not command.wait_reply:
            if command.answer.cancelled():
                return
            command.answer.set_result(cmd_resp)
            return

        ble_notification = await self.ble_get_notification(command.timeout)

        # extract payload from container
        cmd_resp = bytes(ble_notification[1][3:-1])
        if command.answer.cancelled():
            return
        command.answer.set_result(cmd_resp)

    async def protocol_start(self):
        # we assume that every time protocol starts it uses new blank
        # BleakClient to avoid multiple char notifications on every restart
        # bug ?

        assert self.RX_CHAR and self.TX_CHAR
        # if not self.notification_started:
        assert self.client.is_connected
        assert not self._cmd_queue_task.done()
        # check for fresh client
        assert not self.client._notification_callbacks
        _LOGGER.debug(f'Enable BLE notifications from [{self.client.address}]')
        await self.client.write_gatt_char(
            self.TX_CHAR,
            bytearray(0x01.to_bytes(2, byteorder="little")),
            True,
        )
        await self.client.start_notify(
            self.RX_CHAR,
            self.notification_callback,
        )

    async def protocol_stop(self):
        # NB: not used for now as we don't disconnect from our side
        await self.client.stop_notify(self.RX_CHAR)

    async def close(self):
        self.clear_cmd_queue()
        await super().close()

    @staticmethod
    def _check_success(response,
                       error_msg="Command was not completed successfully"):
        success = response and response[0]
        if not success:
            raise RedmondError(error_msg)

    @staticmethod
    def _check_zero_response(response,
                             error_msg="Command was not completed successfully",
                             ):
        response = response and response[0]
        if response != 0:
            raise RedmondError(error_msg)


class RedmondCommonProtocol(RedmondBaseProtocol, abc.ABC):
    """ Shared methods between different devices """
    async def login(self, key):
        _LOGGER.debug('logging in...')
        resp = await self.send_command(Command.AUTH, key, True)
        self._check_success(resp, "Not logged in")

    async def get_version(self):
        _LOGGER.debug('fetching version...')
        resp = await self.send_command(Command.VERSION, b'', True)
        version = tuple(resp)
        _LOGGER.debug(f'version: {version}')
        return version

    async def run(self):
        _LOGGER.debug('Run mode')
        resp = await self.send_command(Command.RUN_CURRENT_MODE)
        self._check_success(resp)

    async def stop(self):
        _LOGGER.debug('Stop mode')
        resp = await self.send_command(Command.STOP_CURRENT_MODE)
        self._check_success(resp)


class RedmondKettle200Protocol(RedmondCommonProtocol, abc.ABC):
    async def set_time(self, ts=None):
        if ts is None:
            ts = time.time()
        ts = int(ts)
        offset = time.timezone \
            if (time.localtime().tm_isdst == 0) else time.altzone
        _LOGGER.debug(f'Setting time ts={ts} offset={offset}')
        resp = await self.send_command(
            Command.SET_TIME,
            struct.pack('<ii', ts, -offset * 60 * 60),
        )
        self._check_zero_response(resp, 'Cannot set time')

    async def get_mode(self):
        _LOGGER.debug('Get mode...')
        response = await self.send_command(Command.READ_MODE)
        return KettleG200State.from_bytes(response)

    async def set_mode(self, state: KettleG200State):
        _LOGGER.debug('Set mode...')
        resp = await self.send_command(
            Command.WRITE_MODE,
            state.to_bytes(),
        )
        success = resp[0]
        if not success:
            raise RedmondError('Cannot set mode')

    async def set_color(self, mode: ColorTarget, r, g, b, brightness):
        # scale_light = [0x00, 0x32, 0x64]
        scale_light = [0x64, 0x64, 0x64]

        rgb_start = rgb_mid = rgb_end = \
            (b << 24) + (g << 16) + (r << 8) + brightness

        resp = await self.send_command(
            Command.WRITE_COLOR,
            struct.pack(
                '<BBIBIBI',
                mode.value,
                scale_light[0],
                rgb_start,
                scale_light[1],
                rgb_mid,
                scale_light[2],
                rgb_end,
            ),
        )
        self._check_zero_response(resp, 'Cannot set color')

    async def get_statistics(self):
        _LOGGER.debug('Get statistics')
        # b'\x00\x00\xdf\x01\x00\x00$\x01\x00\x00\t\x00\x00\x00\x00\x00'
        resp = await self.send_command(Command.GET_STATISTICS, b'\0')
        _, seconds_run, watts_hours, starts, _, _ = \
            struct.unpack('<HIIHHH', resp)
        return {
            'watts_hours': watts_hours,
            'seconds_run': seconds_run,
            'starts': starts,
        }

    async def get_starts_count(self):
        _LOGGER.debug('Get number of start')
        # b'\x00\x00\x00\t\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        # b'\x00\x00\x00\n\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        resp = await self.send_command(Command.GET_STARTS_COUNT, b'\0')
        _, _, starts, *_ = struct.unpack('<BHHHHHHHB', resp)
        return starts


class RedmondCookerProtocol(RedmondCommonProtocol, abc.ABC):
    async def get_mode(self):
        _LOGGER.debug('Get mode...')
        response = await self.send_command(Command.READ_MODE)
        return CookerState.from_bytes(response)

    async def set_mode(self, state: CookerState, ignore_result=True):
        _LOGGER.debug(f'Set mode {state}...')
        resp = await self.send_command(
            Command.WRITE_MODE,
            state.to_bytes(),
        )
        success = resp[0]
        if not ignore_result and not success:
            raise RedmondError('Cannot set mode')

    async def set_predefined_program(self, mode_name: str):
        _LOGGER.debug(f'Set predefined mode {mode_name}...')
        await self.set_mode(COOKER_PREDEFINED_PROGRAMS[mode_name])

    async def set_delay(self, minutes: int):
        _LOGGER.debug('Set delay...')
        resp = await self.send_command(
            Command.WRITE_DELAY,
            bytes([
                minutes // 60,
                minutes % 60,
            ]),
        )
        success = resp[0]
        if not success:
            raise RedmondError('Cannot set delay')

    async def set_temperature(self, temperature: int):
        _LOGGER.debug('Set temperature...')
        resp = await self.send_command(
            Command.WRITE_TEMPERATURE,
            bytes([temperature]),
        )
        success = resp[0]
        if not success:
            raise RedmondError('Cannot set temperature')

    async def set_lock(self, value: bool):
        _LOGGER.debug(f'Set lock {value}...')
        resp = await self.send_command(
            Command.SET_LOCK,
            bytes([1 if value else 0]),
        )
        success = resp[0]
        if not success:
            raise RedmondError('Cannot set lock')

    async def set_sound(self, value: bool):
        _LOGGER.debug(f'Set sound {value}...')
        resp = await self.send_command(
            Command.SET_SOUND,
            bytes([1 if value else 0]),
        )
        success = resp[0]
        if not success:
            raise RedmondError('Cannot set sound')
